/* Copyright (c) 2000-2002 ActiveState Corporation.
   See the file LICENSE.txt for licensing information. */

// The New Plan:
//    - This module is reponsible for MRUs (examples: recent files, recent
//      projects, recent find/replace patterns).
//    - It has the smarts to tie an MRU to a koIOrderedPreference so that
//      it is serialized.
//    - It has the smarts to observe changes in relevant prefs which affect
//      the MRUs, e.g. mruProjectSize.
//      XXX IMO, This is the wrong place to do this. Ideally there would be a
//          koIMRU interface with separate implementations for the various
//          MRUs (and possibly subclassed interfaces for special behaviour).
//    - As required, other modules will call this module's MRU_* methods
//      to note changes in particular MRUs.
//    - This module will send out a "mru_changed" notification whenever an
//      MRU changes. It is up to the various UIs (e.g. File->Recent Files,
//      Getting Started) to observe this notification and act accordingly.
//

//---- globals

try {
    var gMRUPrefObserver = null;
    var mrulog = getLoggingMgr().getLogger('mru');
    //mrulog.setLevel(LOG_DEBUG);
} catch (e) {
    alert(e);
}

//---- internal support routines

function _getMRUListFromOrderedPref(prefName)
{
    var prefSvc = Components.classes["@activestate.com/koPrefService;1"].
                  getService(Components.interfaces.koIPrefService);
    var result = new Array();
    if (prefSvc.prefs.hasPref(prefName)) {
        var mruList = prefSvc.prefs.getPref(prefName);
        for (var i = 0; i < mruList.length; i++) {
            result[i] = mruList.getStringPref(i);
        }
    }
    return result;
}


function _notifyOfMRUChange(prefName)
{
    try {
        var observerSvc = Components.classes["@mozilla.org/observer-service;1"].
                    getService(Components.interfaces.nsIObserverService);
        observerSvc.notifyObservers(null, 'mru_changed', prefName);
    } catch(e) {
        // Raises exception if no observers are registered. Just ignore that.
    }
}


function _MRUPrefObserver ()
{
    this.prefSvc = Components.classes["@activestate.com/koPrefService;1"].
                  getService(Components.interfaces.koIPrefService);
    this.prefSvc.prefs.addObserver(this);
};
_MRUPrefObserver.prototype.destroy = function()
{
    this.prefSvc = Components.classes["@activestate.com/koPrefService;1"].
                  getService(Components.interfaces.koIPrefService);
    this.prefSvc.prefs.removeObserver(this);
    this.prefSvc = null;
}
_MRUPrefObserver.prototype.observe = function(prefSet, prefSetID, prefName)
{
    // Adjust the size of MRUs if the size preference for that MRU changes.
    // XXX Shouldn't really be hardcoding specific MRU names here. This module
    //     is _supposed_ to be generalized.
    if (prefName != "mruProjectSize" && prefName != "mruFileSize"
        && prefName != "mruTemplateSize") {
        return;
    }

    mrulog.info("observed pref '" + prefName + "' change (prefSet=" +
                prefSet + ", prefSetID=" + prefSetID + ")");
    var listPrefName = null;
    if (prefName == "mruProjectSize") {
        listPrefName = "mruProjectList";
    } else if (prefName == "mruFileSize") {
        listPrefName = "mruFileList";
    } else if (prefName == "mruTemplateSize") {
        listPrefName = "mruTemplateList";
    }
    var mruList = null;
    if (this.prefSvc.prefs.hasPref(listPrefName)) {
        mruList = this.prefSvc.prefs.getPref(listPrefName);
    }
    if (mruList) {
        // Chop the list to the correct size.
        var maxEntries = MRU_maxEntries(listPrefName);
        while (mruList.length > maxEntries) {
            mruList.deletePref(mruList.length - 1);
        }
        _notifyOfMRUChange(listPrefName);
    }
};


//---- public methods

function MRU_initialize()
{
    mrulog.info("MRU_initialize()");
    gMRUPrefObserver = new _MRUPrefObserver();
    KomodoRegisterOnUnloadHandler(MRU_finalize);
}


function MRU_finalize()
{
    mrulog.info("MRU_finalize()");
    gMRUPrefObserver.destroy();
    gMRUPrefObserver = null;
}


function MRU_maxEntries(prefName)
{
    mrulog.info("MRU_maxEntries(prefName="+prefName+")");
    // XXX:HACK This should be an attribute on difference instances of a
    //          koIMRU interface.
    var prefSvc = Components.classes["@activestate.com/koPrefService;1"].
                  getService(Components.interfaces.koIPrefService);
    var maxEntries;
    if (prefName == 'mruFileList')  {
        maxEntries = prefSvc.prefs.getLongPref('mruFileSize');
    } else if (prefName == 'mruProjectList')  {
        maxEntries = prefSvc.prefs.getLongPref('mruProjectSize');
    } else if (prefName == 'mruTemplateList')  {
        maxEntries = prefSvc.prefs.getLongPref('mruTemplateSize');
    } else {
        maxEntries = 10;
    }
    return maxEntries;
}


function MRU_add(prefName, entry, caseSensitive)
{
    mrulog.info("MRU_add(prefName="+prefName+", entry="+entry+
                ", caseSensitive="+caseSensitive+")");

    // Add the given "entry" (a string) to the given MRU (indentified by
    // the name of pref, "prefName", with which it is associated).
    // "caseSensitive" is a boolean indicating how to match "entry" with
    // existing MRU entries.
    var prefSvc = Components.classes["@activestate.com/koPrefService;1"].
                  getService(Components.interfaces.koIPrefService);

    // Validate arguments.
    var errmsg;
    if (!prefName) {
        errmsg = "MRU_add: invalid argument: prefName='"+prefName+"'";
        mrulog.error(errmsg);
        throw(errmsg);
    }
    if (!entry) {
        errmsg = "MRU_add: warning: no entry: prefName='"+prefName+
                 "', entry='"+entry+"'";
        mrulog.warn(errmsg)
    }

    var mruList = null;
    if (prefSvc.prefs.hasPref(prefName)) {
        mruList = prefSvc.prefs.getPref(prefName);
    } else {
        // Create a preference-set with this name.
        mruList = Components.classes["@activestate.com/koOrderedPreference;1"]
                  .createInstance();
        mruList.id = prefName;
        // Add the MRU list to the global preference set.
        prefSvc.prefs.setPref(prefName, mruList);
    }

    // If the mru list already contains this entry, first remove it so
    // that it will be reinserted at the top of the list again.
    for (var i = 0; i < mruList.length; i++) {
        var existingEntry = mruList.getStringPref(i);
        if (caseSensitive && entry == existingEntry) {
            mruList.deletePref(i);
            break;
        } else if (entry.toLowerCase() == existingEntry.toLowerCase()) {
            mruList.deletePref(i);
            break;
        }
    }

    // Also: keep the list constrained to the correct size.
    var maxEntries = MRU_maxEntries(prefName);
    while (mruList.length >= maxEntries && mruList.length > 0) {
        mruList.deletePref(mruList.length - 1);
    }

    // "Push" the entry onto the mru "stack".
    if (maxEntries > 0) {
        mruList.insertStringPref(0, entry);
    }

    _notifyOfMRUChange(prefName);
}


function MRU_addURL(prefName, url)
{
    mrulog.info("MRU_addURL(prefName="+prefName+", url="+url+")");

    // Never want to add kodebugger:// URLs to our MRUs.
    // XXX:HACK This should be a special case of a sub-classed implementation
    //          of koIMRU for the recent files and projects MRUs.
    if (url.indexOf("kodebugger://") == 0) {
        return;
    }

    //XXX Might want to cache or preprocess this.
    var infoService = Components.classes['@activestate.com/koInfoService;1'].
                      getService(Components.interfaces.koIInfoService);
    // On *nix filesystems, files with same name different case are recognised
    // as a separate files they are the same file on win32.
    var caseSensitive = infoService.platform.substring(0,3) != "win";

    // max entries should itself come from a preference!
    MRU_add(prefName, url, caseSensitive);
}


function MRU_get(prefName, index /* =0 */)
{
    if (typeof(index) == "undefined" || index == null) index = 0;
    mrulog.info("MRU_get(prefName="+prefName+", index="+index+")");

    var prefSvc = Components.classes["@activestate.com/koPrefService;1"].
                  getService(Components.interfaces.koIPrefService);
    var retval = null;
    if (!prefSvc.prefs.hasPref(prefName)) {
        retval = "";
    } else {
        var mruList = prefSvc.prefs.getPref(prefName);
        if (index >= mruList.length) {
            retval = "";
        } else {
            retval = mruList.getStringPref(index);
        }
    }
    mrulog.info("Returning: " + retval);
    return retval;
}

function MRU_del(prefName, index)
{
    mrulog.info("MRU_del(prefName="+prefName+", index="+index+")");
    // Remove the identified entry from the MRU list.
    var prefSvc = Components.classes["@activestate.com/koPrefService;1"].
                  getService(Components.interfaces.koIPrefService);
    if (prefSvc.prefs.hasPref(prefName)) {
        var mruList = prefSvc.prefs.getPref(prefName);
        mruList.deletePref(index);
        _notifyOfMRUChange(prefName);
    }
}

function MRU_reset(prefName)
{
    mrulog.info("MRU_reset(prefName="+prefName+")");
    // Reset (a.k.a remove all elements) from the given MRU list.
    var prefSvc = Components.classes["@activestate.com/koPrefService;1"].
                  getService(Components.interfaces.koIPrefService);
    if (prefSvc.prefs.hasPref(prefName)) {
        var mruList = prefSvc.prefs.getPref(prefName);
        mruList.reset();
        _notifyOfMRUChange(prefName);
    }
}



//---- MRU Auto-complete support
// Pass the widget and a prefSet name, and magic prevails :)

function MRU_updateFromWidget(widget)
{
    var prefName = widget.getAttribute("komodo-preference-set");
    mrulog.info("MRU_updateFromWidget(widget): widget.value='"+widget.value+
                "', prefName="+prefName);
    MRU_add(prefName, widget.value, true);
}

function MRUAutoCompleteSession(widget)
{
    // This is put inside "if (widget)" because to subclass one needs to run
    // "new MRUAutoCompleteSession();" and then later call this same function
    // with the needed arguments. Thanks a lot JavaScript.
    if (widget) {
        // Hack ourselved into the autocomplete session for the widget
        var prefSetName = widget.getAttribute("komodo-preference-set")
        mrulog.info("MRUAutoCompleteSession(widget): prefName="+prefSetName);
        var sessionName = "mru-" + prefSetName;
        widget.mSessions[sessionName] = this;
        widget.mListeners[sessionName] = new (widget.mAutoCompleteListener)(sessionName);
        ++widget.sessionCount;
        // Do not use <Tab> to cycle through autocomplete list.
        widget.useTabComplete = false;
        this.prefSetName = prefSetName;
    }
}

MRUAutoCompleteSession.prototype = new Object();
MRUAutoCompleteSession.prototype.constructor = MRUAutoCompleteSession;

MRUAutoCompleteSession.prototype.decodeMRUString = function(s)
{
    // A new hook to allow derived classes to decode/demangle the MRU string
    // before use by the user in the autocomplete widget.
    return s;
}

MRUAutoCompleteSession.prototype.onStartLookup = function(value, prevResult, listener)
{
    //dump("***********************  Search Session: onStartLookup\n");
    //dump("***********************  value is:" + value + "\n");
    //dump("***********************  prevResult is:" + prevResult + "\n");
    //dump("onStartLookup has prefs " + _getMRUListFromOrderedPref(this.prefSetName) + "\n");

    var results = Components.classes["@mozilla.org/autocomplete/results;1"]
                  .createInstance();
    results = results.QueryInterface(Components.interfaces.nsIAutoCompleteResults);
    results.searchString = value;
    // no default, i.e. hitting <Return> will use the current string rather
    // than the first match in the MRU
    results.defaultItemIndex = -1;
    results.param = null;

    var mruList = _getMRUListFromOrderedPref(this.prefSetName);
    if (mruList.length == 0) {
        listener.onAutoComplete(results,
            Components.interfaces.nsIAutoCompleteStatus.noMatch);
        return;
    }

    for (var i = 0; i < mruList.length; i++) {
        var thisVal = mruList[i];
        thisVal = this.decodeMRUString(thisVal)
        if (thisVal.indexOf(value)==0) {
            var item = Components.classes["@mozilla.org/autocomplete/item;1"]
                       .createInstance();
            item = item.QueryInterface(Components.interfaces.nsIAutoCompleteItem);
            item.value = thisVal;
            results.items.AppendElement(item);
        }
    }
    listener.onAutoComplete(results,
        Components.interfaces.nsIAutoCompleteStatus.matchFound);
}

MRUAutoCompleteSession.prototype.onStopLookup = function()
{
    return;
}

MRUAutoCompleteSession.prototype.onAutoComplete = function(value, prevResult, listener)
{
    //mrulog.debug("Search Session: onAutoComplete, value =" + value + "\n");
}

