/* Copyright (c) 2000-2003 ActiveState Corporation.
   See the file LICENSE.txt for licensing information. */

/* Functions to convert/parse strings representing URLs, files, etc.
 *
 * This is basically a loose shim around class URIParser in URIlib.py (somewhat
 * obtusely via koIFileEx).
 *
 * Routines:
 *      uriparse_localPathToURI(<localPath>)
 *      uriparse_pathToURI(<URI or localPath>)
 *      uriparse_URIToLocalPath(<URI or localPath>)
 *      uriparse_displayPath(<localPath or URI>)
 *      uriparse_baseName(<localPath or URI>)
 *      uriparse_dirName(<localPath or URI>)
 *      uriparse_ext(<localPath or URI>)
 *      uriparse_fixupURI(<URI>)
 *
 * Dev Notes:
 *  - This module caches a single koIFileEx instance to, presumably, speed
 *    things up.
 */


var _uriparse_koFileEx = null; // koFileEx singleton instance
function _uriparse_getKoFileEx() {
    if (_uriparse_koFileEx == null) {
        _uriparse_koFileEx = Components.classes["@activestate.com/koFileEx;1"]
                             .createInstance(Components.interfaces.koIFileEx);
    }
    _uriparse_koFileEx.URI = null; // clear the instance data
    return _uriparse_koFileEx;
}


// Get the URI representation of the given local file path.
//
//  "localPath" must be a local file path.
//
// Returns the URI for the given path or raises an exception if "localPath" is
// not a local path.
//
// Examples:
//  D:\trentm\foo.txt -> file:///D:/trentm/foo.txt
//  \\planer\d\trentm\tmp\foo.txt -> file://planer/d/trentm/tmp/foo.txt
//  file:///D:/trentm/foo.txt -> throws exception
//  ftp://ftp.activestate.com/ActivePython -> throws exception
//
function uriparse_localPathToURI(localPath) {
    var koFileEx = _uriparse_getKoFileEx();
    koFileEx.path = localPath;
    var os = Components.classes["@activestate.com/koOs;1"].getService();
    if (os.path.normpath(koFileEx.path) != os.path.normpath(localPath)) {
        // ...then this was not a proper local path (probably a URI)
        throw("'"+localPath+"' does not appear to be a proper local path");
    }
    return koFileEx.URI;
}


// Fix the URI using Mozilla's URI fixup service.
//
// Known fixes:
//   - quotes spaces in the URI
//       file:///spam and eggs.txt -> file:///spam%20and%20eggs.txt
//   - adds two leading slashes to file URI with only one
//       file:/foo/bar.txt -> file:///foo/bar.txt
//
// NOTE!!!  nsIURIFixup will escape wide strings incorrectly
//          if it does not understand the uri scheme.  This was
//          discovered for instance, with koremote uri's.  We
//          currently only fixup file:// uri's, but may also need
//          to fixup http:// uri's at some point.
function uriparse_fixupURI(uri) {
    var fixup = Components.classes["@mozilla.org/docshell/urifixup;1"].
                    getService(Components.interfaces.nsIURIFixup);
    try {
        var fixupURI = fixup.createFixupURI(uri,
			Components.interfaces.nsIURIFixup.FIXUP_FLAG_NONE);
            if (fixupURI.spec) {
                var fixeduri = fixupURI.spec;
                if (uri.match(/^file:\/[^\/]/)) {
		    uri = fixeduri;
                }
                // urifixup inserts a '|' after the drive letter; we use colons.
                // So, we have to fix up the fixed-up URL. :-(
                if (fixeduri.match(/^file:\/\/\/.\|\//)) {
                    uri = fixeduri.slice(0, 9) + ":" + fixeduri.slice(10);
                }
                //we also need to fixup UNC paths by hand
                if (fixeduri.match(/^file:\/\/\/\/[^\/]/)) {
                    uri = "file://///" + fixeduri.slice(9);
                }
            }

    } catch (ex) {
        log.debug("nsIURIFixup could not fixup '"+uri+"' - "+ex+"\n");
        // Leave uri alone (presumably to fail below!)
    }
    return uri;
}


// Get the URI representation of the given local file path or URI
//
//  "path" must be a local file path or a URI
//
// Returns the URI for the given path or the URI if one was passed in.
//
// Examples:
//  D:\trentm\foo.txt -> file:///D:/trentm/foo.txt
//  file:///D:/trentm/foo.txt -> file:///D:/trentm/foo.txt
//  ftp://ftp.activestate.com/ActivePython -> ftp://ftp.activestate.com/ActivePython
//
function uriparse_pathToURI(path) {
    var koFileEx = _uriparse_getKoFileEx();
    koFileEx.path = path;
    return koFileEx.URI;
}


// Get the local file path for the given URI.
//
//  "uri" may be a URI for a local file or a local path.
//
// Returns the local file path or raises an exception if there is no local file
// representation for that URI. Note: I would rather this explicitly raised if
// "uri" were a local path, but koIFileEx does not work that way.
//
// Examples:
//  D:\trentm\foo.txt -> D:\trentm\foo.txt
//  file:///D:/trentm/foo.txt -> D:\trentm\foo.txt
//  ftp://ftp.activestate.com/ActivePython -> throws exception
//
function uriparse_URIToLocalPath(uri) {
    var koFileEx = _uriparse_getKoFileEx();
    koFileEx.URI = uri;
    if (koFileEx.scheme != "file") {
        throw("'"+uri+"' does not have a local path");
    }
    return koFileEx.path;
}


// Get an appropriate representation of the given URI for display to the user.
//
//  "uri", typically, is a URI, though it can be a local filename as well.
//
// Examples:
//  file:///D:/trentm/foo.txt -> D:\trentm\foo.txt
//  D:\trentm\foo.txt -> D:\trentm\foo.txt
//  ftp://ftp.activestate.com/ActivePython -> ftp://ftp.activestate.com/ActivePython
//
function uriparse_displayPath(uri) {
    var koFileEx = _uriparse_getKoFileEx();
    koFileEx.URI = uri;
    return koFileEx.displayPath;
}


// Get the basename (a.k.a. leafName) of the given file.
//
//  "file" can be a local filename or URI.
//
// Examples:
//  file:///D:/trentm/foo.txt -> foo.txt
//  D:\trentm\foo.txt -> foo.txt
//  ftp://ftp.activestate.com/ActivePython -> ActivePython
//
function uriparse_baseName(file) {
    var koFileEx = _uriparse_getKoFileEx();
    koFileEx.URI = file;
    return koFileEx.baseName;
}

// Get the dirname of the given file.
//
//  "file" can be a local filename or URI referring to a local file.
//
// Examples:
//  file:///D:/trentm/foo.txt -> D:\trentm
//  D:\trentm\foo.txt -> D:\trentm
//  ftp://ftp.activestate.com/ActivePython -> throws exception
//
function uriparse_dirName(file) {
    var koFileEx = _uriparse_getKoFileEx();
    koFileEx.URI = file;
    if (koFileEx.scheme != "file") {
        throw("'"+file+"' does not have a local dir name");
    }
    return koFileEx.dirName;
}

// Get the extension of the given file.
//
//  "file" can be a local filename or URI
//
function uriparse_ext(file) {
    var koFileEx = _uriparse_getKoFileEx();
    koFileEx.URI = file;
    return koFileEx.ext;
}

