"""
Interface to the Komodo Project Files and their constituent parts.

-- a koIPart is any chunk of data serializable as part of a project file
-- a koIProject is the toplevel koIPart.
-- koIParts can have a prefset attribute which is serialized as a child but available
   as an XPCOM property
"""

from __future__ import generators
import traceback, os, sys, urlparse, string, fnmatch, weakref
from xpcom import components, ServerException, COMException, nsError
import uriparse
from xml.sax import SAXParseException
from xml.sax.saxutils import escape, quoteattr
from xml.dom import pulldom
from koXMLPrefs import NodeToPrefset
from xpcom.server import WrapObject, UnwrapObject
from eollib import newl
import shutil, tempfile
import types
import time
import random
if sys.platform.startswith("win"):
    import pythoncom

KPF_VERSION_STRING = "2"
koLanguageRegistryService = None
gLastProjNum = 1

ANCHOR_MARKER = '!@#_anchor'
CURRENTPOS_MARKER = '!@#_currentPos'

import logging, time
log = logging.getLogger("koProject")

#---- support routines

def _iconRelpathtoURL(relpath, project_dirname):
    # keep this here as part of an upgrade of old style url's for icons
    if relpath.startswith('[PROJECTDIR]'):
        relpath = relpath.replace('[PROJECTDIR]', project_dirname, 1)
        path = os.path.normpath(relpath) # normalize to platform-specific slashes
        url = uriparse.localPathToURI(path)
        return url

    if relpath.startswith('[ICONSDIR]'):
        url = 'chrome://jaguar/content/icons' + relpath[len('[ICONSDIR]'):]
        return url

    return uriparse.localPathToURI(relpath)

def _makeNewProjectName():
    global gLastProjNum
    newname = "Unnamed Project %d" % gLastProjNum
    gLastProjNum += 1
    return newname

def _url2basename(url):
    file = components.classes["@activestate.com/koFileEx;1"].\
            createInstance(components.interfaces.koIFileEx)
    file.URI = url
    return os.path.splitext(file.baseName)[0]


def _genFastGuid():
    if sys.platform.startswith("win"):
        guid = str(pythoncom.CreateGuid())
        guid = guid[1:-1] # strip of the {}'s
    else:
        # this is a hack, but it's good enough for our purposes
        guid = hex(long(time.time()*1000))[2:]
        guid = guid + hex(random.randint(0, sys.maxint-1))[2:] + hex(os.getpid())
    return guid

# we use a weak value dictionary to make sure that the id map doesn't end up
# keeping references alive.
import weakref
_idmap = weakref.WeakValueDictionary()
def getNextId(part):
    xulid = _genFastGuid()
    _idmap[xulid] = part
    return xulid
 
def _findPartById(id):
    if id in _idmap:
        return _idmap[id]
    return None

#---- PyXPCOM implementations

class koPart:
    _com_interfaces_ = [components.interfaces.koIPart]
    _reg_desc_ = "Project Part"
    _reg_contractid_ = "@activestate.com/koPart;1"
    _reg_clsid_ = "{06EB864D-C692-4768-AE45-7900EC5B6E43}"
    keybindable = 0
    type = 'part'
    prettytype = 'Unknown Part'
    _iconurl = ''
    _observerSvc = None
    
    def __init__(self):
        self._prefset = None
        self._attributes = {}
        self.value = ''
        self._tmpAttributes = {} # not serialized
        self._project = None
        self._parent = None
        self.id = getNextId(self)
        self._attributes['id'] = self.id
        self.flavors = ['text/unicode','application/x-komodo-part',\

                        # XXX for a later release, scintilla needs work in this area
                        'TEXT',#,'COMPOUND_TEXT','STRING','UTF-8' \

                        ]
        self.primaryInterface = 'koIPart'

    def assignId(self):
        if self._attributes.has_key('id'):
            self.id = self._attributes['id']
            _idmap[self.id] = self

    def get_keybinding_description(self):
        return self.prettytype + "s: " + self._attributes['name']

    def _getObserverSvc(self):
        if not self._observerSvc:
            self._observerSvc = components.classes["@mozilla.org/observer-service;1"]\
                .getService(components.interfaces.nsIObserverService)
        return self._observerSvc

    def destroy(self):
        self._project = None
        self._parent = None
        self.set_prefset(None)
            
        if hasattr(self, 'children'):
            for child in self.children:
                child = UnwrapObject(child)
                child.destroy()
            del self.children
        #print "refs: ",sys.getrefcount(self)
        
    def get_iconurl(self):
        if self._attributes.has_key('icon'):
            return self._attributes['icon']
        else:
            return self._iconurl

    def _urltorelpath(self, urlin):
        if not self._project:
            return None
        if urlin.startswith('chrome://'):
            return None
        project = UnwrapObject(self._project)
        project_dirname = project._relativeBasedir
        path = uriparse.URIToLocalPath(urlin)
        return uriparse.RelativizeURL(project._relativeBasedir, path)

    def set_iconurl(self, url):
        if not url or url == self._iconurl:
            self.removeAttribute('icon')
            if self._tmpAttributes.has_key('relativeiconurl'):
                del self._tmpAttributes['relativeiconurl']
        else:
            self.setAttribute('icon', url)
            self._tmpAttributes['relativeiconurl'] = self._urltorelpath(url)
        try:
            self._getObserverSvc().notifyObservers(self, 'part_changed', '')
        except Exception, unused:
            pass

    def added(self):
        # we need to walk up the chain of parent containers until we hit
        # either a menu or a toolbar, and notify them that they need to be
        # redrawn
        p = self._parent
        while p and p != p._parent:
            # We need to detect changes to menus and toolbars that already exist
            # the "and p._added" bit is to avoid notifying of changes to menus
            # or toolbars in the deserialization phase (as child parts get
            # their .added() called before their parents.
            if (p.type == 'menu' or p.type == 'toolbar') and p._added:
                try:
                    self._getObserverSvc().notifyObservers(p,
                                                           p.type + '_changed',
                                                           'child added')
                except:
                    pass # no listener
                break # no point in moving any further
            p = p._parent

    def removed(self, previous_parent):
        # we need to walk up the chain of parent containers until we hit
        # either a menu or a toolbar, and notify them that they need to be
        # redrawn
        p = previous_parent
        while p and p != p._parent:
            if p.type == 'menu' or p.type == 'toolbar':
                try:
                    self._getObserverSvc().notifyObservers(p,
                                                           p.type + '_changed',
                                                           'child removed')
                except:
                    pass # no listener
                break # no point in moving any further
            p = p._parent
        
    def __str__(self):
        if self.type in ('file','folder'):
            return 'Part (%s): url=%s, name=%s\nChildren = %s'\
                   % (self.type, getattr(self, 'url', 'NOURL'),
                      getattr(self, 'name', 'NONAME'),
                      getattr(self, 'children', 'No Kids'))
        return 'Part (%s)' % self.type
    
    def get_project(self):
        return self._project
    
    def set_project(self, project):
        self._project = project

    def get_parent(self):
        return self._parent
    
    def set_parent(self, parent):
        self._parent = parent

    def get_prefset(self):
        if not self._prefset:
            prefset = components.classes["@activestate.com/koPreferenceSet;1"] \
                .createInstance(components.interfaces.koIPreferenceSet)
            self.set_prefset(prefset)
            prefset.chainNotifications = 1
        return self._prefset
    
    def set_prefset(self, prefset):
        if self._prefset:
            self._prefset.removeObserver(self)
        self._prefset = prefset
        if self._parent:
            self._prefset.parent = UnwrapObject(self._parent).get_prefset()
        if prefset:
            self._prefset.addObserver(self)

    # prefs observer
    def observe(self, subject, topic, message):
        if self._project:
            project = UnwrapObject(self._project)
            project.set_isDirty(1)
            
    def dump(self, indent):
        print " "*indent + "Part of type '" + self.type +"':"
        for k, v in self._attributes.items():
            print ' '*(indent+1) + "%s: %s" % (k, v)
        print ' '*(indent+1) + 'CDATA: %s' % self.value

    def serialize(self, writer):
        self._serializeHeader(writer)
        self._serializeTail(writer)

    def _serializeHeader(self, writer):
        writer.write("<" + self.type)
        project = UnwrapObject(self._project)
        attrs = dict(self._attributes) # use a copy of the attributes
        if self.type == 'project' and 'url' in attrs:
            del attrs['url']
        elif project._url:
            # use the relative urls from the tmpattributes if th eurl has not changed
            # because RelativizeURL is SLOW
            if 'url' in attrs:
                if 'url' in self._tmpAttributes and self._tmpAttributes['url'] == attrs['url']:
                    attrs['url'] = self._tmpAttributes['relativeurl']
                else:
                    attrs['url'] = uriparse.RelativizeURL(project._relativeBasedir, attrs['url'])







            if self._tmpAttributes.has_key('relativeiconurl'):
                if 'icon' not in attrs:
                    attrs['icon'] = self._tmpAttributes['relativeiconurl']
                else:
                    attrs['icon'] = uriparse.RelativizeURL(project._relativeBasedir, attrs['icon'])
        for a_name, data in attrs.items():
            if type(data) not in types.StringTypes:
                data = str(data)
            if '\n' in data:
                errmsg = "Illegal newline character in '%s' attribute of "\
                         "%s project item" % (a_name, self.type)
                lastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"]\
                    .getService(components.interfaces.koILastErrorService)
                lastErrorSvc.setLastError(0, errmsg)
                raise ServerException(nsError.NS_ERROR_ILLEGAL_VALUE, errmsg)
            # XXX need to decentralize this processing
            writer.write(" %s=%s" % (a_name, quoteattr(data)))
        writer.write(">%s"%(newl))
        if self._prefset:
            # unwrap so we can use cStringIO
            prefset = UnwrapObject(self._prefset)
            prefset.serialize(writer)
        # write out the CDATA, i.e. self.value
        writer.write(escape(self.value))

    def _serializeTail(self, writer):
        writer.write("</%s>%s" % (self.type,newl))

    def hasAttribute(self, name):
        return self._attributes.has_key(name)

    def getAttribute(self, name):
        return self._attributes[name]
    
    def setAttribute(self, name, value):
        if name not in self._attributes or self._attributes[name] != value: # avoid dirtification when possible.
            self._attributes[name] = value
            if self._project:
                UnwrapObject(self._project).set_isDirty(1)

    def removeAttribute(self, name):
        if name not in self._attributes: return
        del self._attributes[name]
        if self._project:
            UnwrapObject(self._project).set_isDirty(1)

    def getStringAttribute(self, name):
        return unicode(self._attributes[name])
    
    def setStringAttribute(self, name, value):
        self.setAttribute(name, unicode(value))

    def getLongAttribute(self, name):
        return int(self._attributes[name])
    
    def setLongAttribute(self, name, value):
        self.setAttribute(name, int(value))

    def getBooleanAttribute(self, name):
        return self.getLongAttribute(name) and 1 or 0

    def setBooleanAttribute(self, name, value):
        self.setAttribute(name, value and 1 or 0)

    def getDragData(self):
        #print "getDragData ",repr(self.value)
        return self.value
    
    def getDragFlavors(self):
        return self.flavors        
    
    def clone(self):
        # This clone function handles only the basic information that is
        # in every koIPart.  Custom parts that have additional information
        # which needs to be cloned will need to implement their own clone
        # function.  One small example is koIContainer
        kopart = components.classes[self._reg_contractid_] \
              .createInstance(self._com_interfaces_[0])
        part = UnwrapObject(kopart)
        part.type = self.type
        part.value = self.value

        if hasattr(self, '_url') and part.type != 'project':  # we _don't_ want clones of projects to have a URL
            part._url = self._url
        if hasattr(self, '_name'):
            part._name = self._name

        if self._prefset:
            part.set_prefset(self._prefset.clone())

        a_names = self._attributes.keys()
        a_names.sort()
        for a_name in a_names:
            data = self.getStringAttribute(a_name)
            part.setStringAttribute(a_name,data)
            
        # this *MUST* be last in the clone process.  If it is not
        # then a clone will dirty the project that contains the item
        # being cloned.
        if self._project == self: # deal with project toplevel cycle
            part._project = part
        else:
            part._project = self._project

        return WrapObject(part, self._com_interfaces_[0])
    
    def applyKeybindings(self, recurse):
        if self._attributes.get('keyboard_shortcut', ''):
            try:
                self._getObserverSvc().notifyObservers(self, 'kb-load', self.id)
            except Exception, unused:
                pass
        if recurse and hasattr(self, "children"):
            for child in self.children:
                child.applyKeybindings(recurse)

    def invoke(self):
        try:
            self._getObserverSvc().notifyObservers(self,
                                                   'part-invoke',
                                                   self.id)
        except Exception, unused:
            pass

class koContainerBase(koPart):
    def __init__(self):
        koPart.__init__(self)
        self.childrenwrapped = 1
        self._name = ''
        self._url = ''
        self.numChildren = 0
        self.children = []

    def getChildren(self):
        # Not sure if this is the right time to do XPCOM wrapping.
        if not self.childrenwrapped:
            self._wrapChildren()
        return self.children
    
    def _wrapChildren(self):
        self.childrenwrapped = 1
        self.children = [WrapObject(child, components.interfaces.koIPart)\
                         for child in self.children]

    def hasChild(self, child):
        return child in self.children

    def isEmpty(self):
        return len(self.children)==0
    
    def getChildrenByType(self, type, recurse):
        # This optimization assumes that all koPart implementations are
        # Python objects.  Otherwise we'd need to use getAttribute,
        # which would be much slower.
        if self.childrenwrapped:
            children = [UnwrapObject(child) for child in self.children]
        else:
            children = self.children
        matching_children = []
        
        for child in children:
            if child.type == type:
                matching_children.append(child)
            elif recurse and hasattr(child, 'children'):
                matching_children += child.getChildrenByType(type, recurse)
        return matching_children
    
    def getChildWithTypeAndStringAttribute(self, type, attrname, attrvalue, recurse):
        # This optimization assumes that all koPart implementations are
        # Python objects.  Otherwise we'd need to use getAttribute,
        # which would be much slower.
        if self.childrenwrapped:
            children = [UnwrapObject(child) for child in self.children]
        else:
            children = self.children
        for child in children:
            if child.type == type\
               and child._attributes.get(attrname,None) == attrvalue:
                return child
            elif recurse and hasattr(child, 'children'):
                c = child.getChildWithTypeAndStringAttribute(type,attrname,attrvalue,recurse)
                if c: return c
        return None

    def getChildByAttributeValue(self, attrname, attrvalue, recurse):
        if self.childrenwrapped:
            children = [UnwrapObject(child) for child in self.children]
        else:
            children = self.children
        for child in children:
            if child._attributes.get(attrname,None) == attrvalue:
                return child
            elif recurse and hasattr(child, 'children'):
                c = child.getChildByAttributeValue(attrname,attrvalue,recurse)
                if c: return c
        return None

    def addChild(self, child):
        if not self.childrenwrapped:
            self._wrapChildren()
        if not self.hasChild(child):
            self.children.append(child)
        unwrapped_child = UnwrapObject(child)
        unwrapped_child._parent = self
        if unwrapped_child._prefset:
            unwrapped_child._prefset.parent = self.get_prefset()
        if self._project:
            project = UnwrapObject(self._project)
            project.set_isDirty(1)
            self._project.registerChildByURL(child)
            child.project = self._project # this calls the setter
            if child.type == "file":
                # Let the Code Intelligence system know that a file has been
                # added to one of Komodo's open projects or toolboxes.
                self._getCodeIntelSvc().ideEvent("added_file_to_project",
                                                 child.getStringAttribute("url"),
                                                 None)
            
        child.added()

    _codeIntelSvc = None
    def _getCodeIntelSvc(self):
        if not self._codeIntelSvc:
            self._codeIntelSvc = components.classes["@activestate.com/koCodeIntelService;1"]\
                .getService(components.interfaces.koICodeIntelService);
        return self._codeIntelSvc
    
    def getChildById(self, id):
        if self.id == id:
            return self
        for child in self.children:
            if child.id == id:
                return child
            try:
                child = child.QueryInterface(components.interfaces.koIContainerBase)
            except xpcom.nsError:
                # QI failed
                continue
            found = child.getChildById(id)
            if found: return found
        return None
        
    def get_project(self):
        return self._project
    
    def set_project(self, project):
        self._project = project
        # Need to recursively go through the child's children to set the
        # project property.
        for child in self.children:
            child.project = project# This is recursive 

    def removeChild(self, child):
        if self.childrenwrapped:
            children = [UnwrapObject(c) for c in self.children]
        child = UnwrapObject(child)
        children.remove(child)
        if self._project:
            UnwrapObject(self._project).set_isDirty(1)
            self._project.forgetChildByURL(child)
            if child.type == "file":
                # Let the Code Intelligence system know that a file has been
                # removed from one of Komodo's open projects or toolboxes.
                self._getCodeIntelSvc().ideEvent("removed_file_from_project",
                                                 child.getStringAttribute("url"),
                                                 None)

        if child.keybindable:
            try:
                self._getObserverSvc().notifyObservers(child, 'kb-unload', child.id)
            except Exception, unused:
                pass

        previous_parent = child._parent
        child.project = None
        child._parent = None
        self.children = children
        self._wrapChildren()
        child.removed(previous_parent)

    def removed(self, my_previous_parent):
        # now we need to look at our children, and remove any menu's or toolbars
        needRemoved = [UnwrapObject(child) for child in \
                       self.getChildrenByType('toolbar', 1) + \
                       self.getChildrenByType('menu', 1)]
        for child in needRemoved:
            child.removed(child._parent)
        
    def serialize(self, writer):
        # just like parts, except there are children to consider.
        self._serializeHeader(writer)
        self._serializeChildren(writer)
        self._serializeTail(writer)

    def _serializeChildren(self, writer):
        if self.childrenwrapped:
            children = [UnwrapObject(child) for child in self.children]
        else:
            children = self.children

        for part in children:
            part.serialize(writer)

    def set_name(self, name):
        self._name = name
        if hasattr(self, '_url') and self._url:
            # using lists to make it mutable
            parts = list(urlparse.urlparse(self._url))
            slashlocation = parts[2].rindex('/')
            parts[2] = parts[2][:slashlocation+1] + name
            if parts[2].find('.kpf') < 0:
                parts[2] = parts[2] + '.kpf'
            self._url = urlparse.urlunparse(tuple(parts))
        self.setStringAttribute('name', self._name)
        if self._project:
            UnwrapObject(self._project).set_isDirty(1)

    def set_url(self, url):
        # XXX should do URL validation
        if url.find('://') == -1:
            url = uriparse.localPathToURI(url)
        self._url = url
        self.setStringAttribute('url', self._url)
        # Need to do the same checks as set_name.
        self._name = _url2basename(url)
        self._relativeBasedir = self._url[:self._url.rindex('/')+1]
        self.setStringAttribute('name', self._name)
        if self._project:
            UnwrapObject(self._project).set_isDirty(1)

    def get_url(self):
        if not self._url:
            self._url = self.getStringAttribute('url')
        return self._url

    def get_name(self):
        return self._name
    
    def clone(self):
        part = koPart.clone(self)
        for child in self.children:
            part.addChild(child.clone())
        return part
        
    def dump(self, indent):
        koPart.dump(self, indent)
        for child in self.children:
            child.dump(indent+2)

class koContainer(koContainerBase):
    def __init__(self):
        koContainerBase.__init__(self)





















# See factory function below
class koFilePart(koPart):
    _com_interfaces_ = [components.interfaces.koIPart_file,



                       ]
    _reg_desc_ = "Komodo File"
    _reg_contractid_ = "@activestate.com/koPart?type=file;1"
    _reg_clsid_ = "{8FD46D97-B2F9-4875-8D36-82BFD7471A97}"
    type = 'file'
    prettytype = 'File'
    _iconurl = 'chrome://jaguar/skin/images/file_icon.png'
    
    def __init__(self):
        koPart.__init__(self)
        self.flavors.insert(0,'text/x-moz-url')
        self.primaryInterface = 'koIPart_file'
        
    def __str__(self):
        return "<koPart_file, url=%s>" % self._attributes.get('url', 'None')

    def getDragData(self):
        return self.getStringAttribute('url')

def File(url, name, project=None):
    """Construct a 'file' koIPart."""
    part = koFilePart()
    part._attributes['url'] = url
    part._attributes['name'] = name
    part._project = project
    return part

def ProjectShortcut(url, name, project=None):
    """Construct a 'file' koIPart."""
    part = koProjectRef()
    part._attributes['url'] = url
    part._attributes['name'] = name
    part._project = project
    return part











class koSnippetPart(koPart):
    _com_interfaces_ = [components.interfaces.koIPart_snippet]
    _reg_desc_ = "Komodo Snippet"
    _reg_contractid_ = "@activestate.com/koPart?type=snippet;1"
    _reg_clsid_ = "{D23CF82A-7E80-470d-8C6F-026193EF07F7}"
    type = 'snippet'
    prettytype = 'Snippet'
    _iconurl = 'chrome://jaguar/skin/images/snippet.gif'

    def __init__(self):
        koPart.__init__(self)
        self.keybindable = 1
        self.primaryInterface = 'koIPart_snippet'

    def getDragData(self):
        eol = '\n'
        # NOTE: IT IS IMPORTANT THAT IF UNICODE COMES IN, UNICODE GOES OUT!
        newtext = u'' # THE u IS IMPORTANT!
        text = self.value
        i = 0
        while i < len(text):
            if text[i] == '\\':
                i += 1
                if text[i] == 'n':
                    newtext += eol
                elif text[i] == 't':
                    newtext += '\t'
                elif text[i] == '\\':
                    newtext += '\\'
                else:
                    i -= 1
                    newtext == '\\'
            else:
                newtext += text[i]
            i += 1

        anchor = newtext.find(ANCHOR_MARKER)
        if anchor != -1:
            newtext = newtext[:anchor] + newtext[anchor+len(ANCHOR_MARKER):]
        currentPos = newtext.find(CURRENTPOS_MARKER)
        if currentPos != -1:
            newtext = newtext[:currentPos] + newtext[currentPos+len(CURRENTPOS_MARKER):]
        # it would be good to interpolate here as well!
        gISvc = components.classes["@activestate.com/koInterpolationService;1"]\
                .getService(components.interfaces.koIInterpolationService);
        try:
            queries, strings = gISvc.Interpolate1([], # not bracketed
                                                  [newtext], # bracketed
                                                  None,  #filename
                                                  0, #linenum
                                                  None, #word
                                                  None, #selection
                                                  None, #prefSet: use global prefs
                                                  )
        except: # If e.g. %w is in the snippet, we ignore all % codes.
            return newtext
        
        for query in queries:
            if query.answer == None:
                # if any of the interpolations required a question,
                # do _no_ conversion and return the full %-laden snippet
                return newtext
        if queries:
            strings = gISvc.Interpolate2(strings, queries)
        
        return strings[0]

class koCommandPart(koPart):
    _com_interfaces_ = [components.interfaces.koIPart_command]
    _reg_desc_ = "Komodo Command"
    _reg_contractid_ = "@activestate.com/koPart?type=command;1"
    _reg_clsid_ = "{AC02EC02-08B2-4c87-9861-D9917144A856}"
    type = 'command'
    prettytype = 'Run Command'
    _iconurl = 'chrome://jaguar/skin/images/run_commands.png'

    def __init__(self):
        koPart.__init__(self)
        self.keybindable = 1
        self.primaryInterface = 'koIPart_command'

class koTemplatePart(koPart):
    _com_interfaces_ = [components.interfaces.koIPart_template]
    _reg_desc_ = "Komodo Template"
    _reg_contractid_ = "@activestate.com/koPart?type=template;1"
    _reg_clsid_ = "{473BD963-632C-403e-953F-E7F2F2CA84A1}"
    type = 'template'
    prettytype = 'Template'
    _iconurl = 'chrome://jaguar/skin/images/newTemplate.png'
    
    def __init__(self):
        koPart.__init__(self)
        self.primaryInterface = 'koIPart_template'

class koURLPart(koPart):
    _com_interfaces_ = [components.interfaces.koIPart_URL]
    _reg_desc_ = "URL"
    _reg_contractid_ = "@activestate.com/koPart?type=URL;1"
    _reg_clsid_ = "{497C73A6-9405-4d0f-A472-8C7343AC5D60}"
    type = 'URL'
    prettytype = 'URL'
    _iconurl = 'chrome://jaguar/skin/images/xlink.gif'
    keybindable = 1

    def __init__(self):
        koPart.__init__(self)
        self.primaryInterface = 'koIPart_URL'

class koMacroPart(koPart):
    _com_interfaces_ = [components.interfaces.koIPart_macro]
    _reg_desc_ = "Macro"
    _reg_contractid_ = "@activestate.com/koPart?type=macro;1"
    _reg_clsid_ = "{4884D548-698E-45e2-9324-715F623CD300}"
    type = 'macro'
    prettytype = 'Macro'
    _iconurl = 'chrome://jaguar/skin/images/macro.png'
    
    def __init__(self):
        koPart.__init__(self)
        self.flavors.insert(0,'text/x-moz-url')
        self.keybindable = 1
        self.primaryInterface = 'koIPart_macro'
        self.setStringAttribute('language', 'JavaScript')  # default.

    def added(self):
        # if we're an old-style macro, convert us to new style
        if self.hasAttribute('filename'):
            # need to upgrade
            fname = self.getAttribute('filename')
            if not os.path.exists(fname):
                log.warn("Couldn't find the file %r referred to by a macro", fname)
                return
            registryService = components.classes['@activestate.com/koLanguageRegistryService;1'].\
                                         getService(components.interfaces.koILanguageRegistryService)
            language = registryService.suggestLanguageForFile(os.path.basename(fname))
            try:
                file = open(fname, 'r')
                self.value = file.read()
                file.close()
                self.setAttribute('language', language)
                self.setAttribute('version', '2')
                self.setBooleanAttribute('do_trigger', 0)
                self.setLongAttribute('rank', 100)
                self.setBooleanAttribute('trigger', 'trigger_presave')
                self.removeAttribute('filename') # ADD when removeAttribute is added
            except OSError:
                log.warn("Couldn't read the macro: %r, skipping conversion", fname)
                return
        try:
            self._getObserverSvc().notifyObservers(self,'macro-load','')
        except:
            pass # no listener
        koPart.added(self)

    def removed(self, previous_parent):
        try:
            self._getObserverSvc().notifyObservers(self, 'macro-unload', '')
        except:
            pass # no listener
        koPart.removed(self, previous_parent)

    def evalAsPython(self, domdocument, window, scimoz, document,
                     view, code, async):
        if async:
            lastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"]\
                .getService(components.interfaces.koILastErrorService)
            lastErrorSvc.setLastError(1, "Asynchronous python macros not yet implemented")
            raise ServerException
        evalPythonMacro(self, domdocument, window, scimoz, document, view, code)

# See factory functions below
class koWebServicePart(koURLPart):
    _com_interfaces_ = [components.interfaces.koIPart_URL,
                        components.interfaces.koIPart_webservice]
    _reg_desc_ = "Komodo Web Service"
    _reg_contractid_ = "@activestate.com/koPart?type=webservice;1"
    _reg_clsid_ = "{00344B63-8DA5-4145-9D87-ABD13190FDC2}"
    type = 'webservice'
    prettytype = 'Web Service Shortcut'
    _iconurl = 'chrome://jaguar/skin/images/webServicesImg.png'
    
    def __init__(self):
        koPart.__init__(self)
        self.setStringAttribute('wsdl','');
        self.primaryInterface = 'koIPart_URL'

    def hasAttribute(self, name):
        if name.lower() == 'url':
            return 'wsdl' in self._attributes
        return name in self._attributes

    def get_value(self):
        return unicode(self._attributes['wsdl'])
    
    def getStringAttribute(self, name):
        if name.lower() == 'url':
            return unicode(self._attributes['wsdl'])
        return unicode(self._attributes[name])

    def setStringAttribute(self, name, value):
        if name.lower() == 'url':
            name = 'wsdl'
        self.setAttribute(name, unicode(value))

    def getDragData(self):
        return self.getStringAttribute('wsdl')


class koMenuPart(koContainer):
    _com_interfaces_ = [components.interfaces.koIPart_menu]
    _reg_desc_ = "Komodo Menu"
    _reg_contractid_ = "@activestate.com/koPart?type=menu;1"
    _reg_clsid_ = "{2FCC515F-F2B8-4DB4-B2BB-664615273275}"
    type = 'menu'
    prettytype = 'Custom Menu'
    _iconurl = 'chrome://jaguar/skin/images/menu_icon.png'
    _added = 0
    def __init__(self):
        koContainer.__init__(self)
        self.setStringAttribute('accesskey', '')
        self.setLongAttribute('priority', 100)
        self._added = 0

    def setAttribute(self, name, value):
        koContainer.setAttribute(self, name, value)
        if self._added and not UnwrapObject(self._project)._quiet:
            try:
                self._getObserverSvc().notifyObservers(self, 'menu_changed', 'attribute changed')
            except:
                pass # no listener

    def added(self):
        if UnwrapObject(self._project)._quiet:
            return
        try:
            self._getObserverSvc().notifyObservers(self, 'menu_create', 'menu_create')
        except:
            pass # no listener
        koContainer.added(self)
        self._added = 1

    def removed(self, previous_parent):
        if UnwrapObject(self._project)._quiet:
            return
        if not self._added:
            return
        self._added = 0
        try:
            self._getObserverSvc().notifyObservers(self, 'menu_remove', 'menu_remove')
        except:
            pass # no listener
        koContainer.removed(self, previous_parent)
        
    def getDragData(self):
        return self.getStringAttribute('name')
    

class koToolbarPart(koContainer):
    _com_interfaces_ = [components.interfaces.koIPart_toolbar]
    _reg_desc_ = "Komodo Toolbar"
    _reg_contractid_ = "@activestate.com/koPart?type=toolbar;1"
    _reg_clsid_ = "{71A11015-F116-4466-9956-7D6859B64CB4}"
    type = 'toolbar'
    prettytype = 'Custom Toolbar'
    _iconurl = 'chrome://jaguar/skin/images/toolbar_icon.png'
    def __init__(self):
        koContainer.__init__(self)
        self._added = 0
        self.setLongAttribute('priority', 100)

    def setAttribute(self, name, value):
        koContainer.setAttribute(self, name, value)
        if self._added and not UnwrapObject(self._project)._quiet:
            try:
                self._getObserverSvc().notifyObservers(self, 'toolbar_changed', 'attribute changed')
            except:
                pass # no listener

    def added(self):
        if UnwrapObject(self._project)._quiet:
            return
        self._added = 1
        # we've been created, we need to tell the JS side of things
        # that we exist.
        try:
            self._getObserverSvc().notifyObservers(self, 'toolbar_create', 'toolbar_create')
        except:
            pass # no listener
        koContainer.added(self)

    def removed(self, previous_parent):
        if UnwrapObject(self._project)._quiet:
            return
        if not self._added:
            return
        self._added = 0
        try:
            self._getObserverSvc().notifyObservers(self, 'toolbar_remove', 'toolbar_remove')
        except:
            pass # no listener
        koContainer.removed(self, previous_parent)

    def getDragData(self):
        return self.getStringAttribute('name')


# See factory functions below
class koFolderPart(koContainer):
    _com_interfaces_ = [components.interfaces.koIPart_folder,



                       ]
    _reg_desc_ = "Komodo Folder"
    _reg_contractid_ = "@activestate.com/koPart?type=folder;1"
    _reg_clsid_ = "{7E774CEB-07F5-40da-AC37-954779B2F09F}"
    type = 'folder'
    prettytype = 'Folder'
    _iconurl = 'chrome://jaguar/skin/images/folder-closed.png'

    def __init__(self):
        koContainer.__init__(self)
        self.primaryInterface = 'koIPart_folder'

    def getLanguageFolder(self, language):
        for child in self.children:
            if child.hasAttribute('language') and child.getStringAttribute('language') == language:
                return child
        languagefolder = Folder('', language, self._project)
        languagefolder.setStringAttribute('language', language)
        self.addChild(languagefolder)
        return languagefolder

    def getDirFolder(self, basedir, targetdir):
        # we want to get (or create, if necessary) the folder structure that corresponds to the
        # difference between basedir and targetdir.
        # e.g. if basedir is c:/foo/bar, and targetdir is c:/foo/bar/baz/fob
        # we want to return a fob Folder inside a bar Folder, which itself would be added to self.
        if not targetdir.startswith(basedir):
            log.error("getDirFolder called with invalid arguments: %s, %s" % (basedir, targetdir))
            return None
        trail = targetdir[len(basedir):]
        folder = self
        if trail:
            path = basedir
            parts = trail.split(os.sep)
            for part in parts:
                if not part: continue
                path = os.path.join(path, part)
                url = uriparse.localPathToURI(path)
                for child in folder.children:
                    if child.hasAttribute('url') and child.getStringAttribute('url') == url:
                        folder = child
                        break
                else:
                    newfolder = Folder(url, part, self._project)
                    folder.addChild(newfolder)
                    folder = newfolder
        return folder
            

    def getDragData(self):
        if self.hasAttribute('url'):
            return self.getStringAttribute('url');
        return self.getStringAttribute('name');
    
    def getDragFlavors(self):
        if self.hasAttribute('url') and self.flavors[0] != 'text/x-moz-url':
            self.flavors.insert(0,'text/x-moz-url')
        return self.flavors        

def Folder(url, name, project=None):
    """Construct a koIPart_folder from URL and name."""
    part = koFolderPart()
    part._attributes['url'] = url
    part._attributes['name'] = name
    part._project = project
    return part

class koProjectRef(koFilePart):
    _com_interfaces_ = [components.interfaces.koIPart_ProjectRef]
    _reg_desc_ = "Project Reference"
    _reg_contractid_ = "@activestate.com/koPart?type=ProjectRef;1"
    _reg_clsid_ = "{0A027B8B-43B0-4da6-B26E-0426EA803208}"
    type = 'ProjectRef'
    prettytype = 'Project Shortcut'
    _iconurl = 'chrome://jaguar/skin/images/project_icon.png'
    
    def __init__(self):
        koFilePart.__init__(self)
        self.primaryInterface = 'koIPart_ProjectRef'

class koDirectoryShortcut(koPart):
    _com_interfaces_ = [components.interfaces.koIDirectoryShortcut]
    _reg_desc_ = "Directory Shortcut"
    _reg_contractid_ = "@activestate.com/koPart?type=DirectoryShortcut;1"
    _reg_clsid_ = "{ED80A695-BF4F-4BFE-8857-74177ADA7246}"
    type = 'DirectoryShortcut'
    prettytype = 'Open... Shortcut'
    keybindable = 1
    _iconurl = 'chrome://jaguar/skin/images/open.png'
    
    def __init__(self):
        koPart.__init__(self)
        self.primaryInterface = 'koIDirectoryShortcut'

class koToolbox(koPart):
    _com_interfaces_ = [components.interfaces.koIToolbox]
    _reg_desc_ = "Toolbox"
    _reg_contractid_ = "@activestate.com/koPart?type=Toolbox;1"
    _reg_clsid_ = "{D1438011-23EA-491A-9601-7AD5729A5369}"
    type = 'Toolbox'

class koUnknown(koPart):
    _com_interfaces_ = [components.interfaces.koIPart]
    _reg_desc_ = "Unknown"
    _reg_contractid_ = "@activestate.com/koPart?type=Unknown;1"
    _reg_clsid_ = "{d751757b-9a7a-449e-b756-93ce24e39acd}"
    type = 'Unknown'

class koProject(koFolderPart):
    _com_interfaces_ = [components.interfaces.koIProject,




                       ]
    _reg_desc_ = "Komodo Project"
    _reg_contractid_ = "@activestate.com/koProject;1"
    _reg_clsid_ = "{39EE799F-D0F5-4b2d-A5A6-34151B2AC235}"
    type = 'project'
    prettytype = 'Project'
    _iconurl = 'chrome://jaguar/skin/images/project_icon.png'
    
    def __init__(self):
        koFolderPart.__init__(self)
        self._isDirty = 0
        self._urlmap = weakref.WeakValueDictionary()
        self._project = self # cycle - need to investigate WeakReference
        self.primaryInterface = 'koIProject'
        self.lastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"]\
                       .getService(components.interfaces.koILastErrorService)
        self._prefset = components.classes["@activestate.com/koPreferenceSet;1"] \
                        .createInstance(components.interfaces.koIPreferenceSet)
        self._prefset.chainNotifications = 1
        # this is the toplevel project, parent prefs are the global prefs
        self._prefset.parent = components.classes["@activestate.com/koPrefService;1"].\
                                getService(components.interfaces.koIPrefService).prefs
        self._quiet = 0

    # Fill with a new empty, unnamed project 
    def create(self):
        self._name = _makeNewProjectName()
        self.setStringAttribute("kpf_version", KPF_VERSION_STRING)
    
    def _get_stream(self, path, mode):  # mode is 'r' or 'w'
        self.name = os.path.splitext(os.path.basename(path))[0]
        stream = components.classes["@activestate.com/koLocalFile;1"].\
             createInstance().QueryInterface(components.interfaces.koILocalFile)
        try:
            stream.init(path, mode)
        except (IOError, COMException), e:
            errstr = "Can't open project file %s because %s" % (path, e)
            log.error(errstr)
            raise 
        return stream

    def set_isDirty(self, dirty):
        if self._isDirty != dirty:
            self._isDirty = dirty
            try:
                self._getObserverSvc().notifyObservers(self,'update-commands','projectsChanged')
            except:
                pass # no listener
    
    def get_isDirty(self):
        return self._isDirty
    
    def loadQuiet(self, url):
        self._quiet = 1
        self.load(url)
        self._quiet = 0
        
    def load(self, url):
        t1 = time.clock()
        self.lastErrorSvc.setLastError(0, '')
        self.set_url(url)
        assert self._url
        assert self._relativeBasedir
        fname = uriparse.URIToLocalPath(self._url)
        if not os.path.exists(fname):
            self.lastErrorSvc.setLastError(0, 'File does not exist.')
            raise ServerException
            
        stream = self._get_stream(fname, 'r')
        self.children=[]
        
        # part is our 'current part' that we are working on
        # we need to have this so we can add prefsets to them
        # as they pass by
        part = self
        # partstack is a list of parts by depth
        partstack = []
        # our current level of depth
        level = 0

        dirtyAtEnd = 0
        events = pulldom.parse(stream)

        keybound_parts = []  # list of parts that have keybindings
        for (event, node) in events:
            if event == pulldom.START_ELEMENT:
                if node.tagName == 'project':
                    # assuming this is hapening in koIProject we need to
                    # add attributes to self
                    for attribute, value in node.attributes.items():
                        self._attributes[attribute] = value
                    partstack.append(self)

                elif node.tagName == "preference-set":
                    # make a dom for the prefset, turn it into a prefset
                    events.expandNode(node)
                    part.set_prefset(NodeToPrefset(node))
                    if self == part:
                        # this is the toplevel project, parent prefs are the global prefs
                        self._prefset.parent = components.classes["@activestate.com/koPrefService;1"].\
                                getService(components.interfaces.koIPrefService).prefs


                elif node.tagName == 'files':
                    # ignore the obsolete 'files' nodes, we'll grab the children later on
                    continue

                elif node.tagName == 'languagefolder':
                    # convert old 'languagefolder' types to regular folders
                    part = koFolderPart()
                    language = node.attributes['language'].value
                    part._attributes['name'] = language
                    part._attributes['language'] = language
                    part.value = ''
                    part._project = self
                    partstack.append(part)
                    continue

                else:
                    # create our new part instance
                    try:
                        part = createPartFromType(node.tagName)
                    except KeyError, e:
                        # a bad part in the project, ignore for now
                        log.error("Unknown project element with name %s", node.tagName)
                        part = koUnknown()
                    
                    for attribute, value in node.attributes.items():
                        part._attributes[attribute] = value

                    # XXX need to decentralize this processing
                    if 'url' in part._attributes:
                        value = part._attributes['url']
                        if value.find('://') == -1: # it is not an real URL -- it must be relative
                            part._tmpAttributes['relativeurl'] = value
                            value = uriparse.UnRelativizeURL(self._relativeBasedir, value)
                            part._attributes['url'] = part._tmpAttributes['url'] = value
                        self._urlmap[value] = part
                    if 'iconrelpath' in part._attributes:
                        # fixup our old icon paths
                        value = part._attributes['iconrelpath']
                        if value[0] in ['/','['] or value[1] == ':':
                            # pre 3.0 project icon url's, fixem now, and mark the project
                            # as dirty so it needs to be saved with corrected urls
                            value = _iconRelpathtoURL(value, self._relativeBasedir)
                            relpath = uriparse.RelativizeURL(self._relativeBasedir, value)
                            part._attributes['icon'] = relpath
                        else:
                            part._attributes['icon'] = value
                        del part._attributes['iconrelpath']
                        dirtyAtEnd = 1
                    if 'icon' in part._attributes:
                        value = part._attributes['icon']
                        if value.find('://') == -1: # it is not an real URL -- it must be relative
                            part._tmpAttributes['relativeiconurl'] = value
                            value = uriparse.UnRelativizeURL(self._relativeBasedir, value)
                        part._attributes['icon'] = value
                    if 'startup' in part._attributes and \
                       part._attributes['startup'].find('://') == -1:
                            value = part._attributes['startup']
                            part._tmpAttributes['relativestartup'] = value
                            part._attributes['startup'] = part._tmpAttributes['startup'] = uriparse.UnRelativizeURL(self._relativeBasedir, value)
                    if 'keyboard_shortcut' in part._attributes and \
                       part._attributes['keyboard_shortcut']:
                        keybound_parts.append(part)
                    
                    part.assignId() # XXX why here and in END_ELEMENT below?
                    part.value = ''
                    part._project = self
                    # push the part onto a stack
                    partstack.append(part)
                
            elif event == pulldom.END_ELEMENT and \
               not node.tagName in ['files','project','preference-set']:
                # pop a part off the stack
                childpart = partstack.pop()

                if childpart.type == 'snippet':
                    # "old" (2.0 and 2.3b1) snippets need to be converted to the new
                    # encoding scheme.  New snippets are fine the way they are.
                    childpart = _SnippetConvert(childpart)
        
                # then make it a child of the parent
                # if the parent is only a koIPart, then we
                # get an exception, it's an invalid project
                part = partstack[-1]
                childpart._parent = part
                if childpart._prefset:
                    childpart._prefset.parent = part.get_prefset()
                childpart.assignId()
                childpart._project = self
                part.children.append(childpart)
                childpart.added()

            elif event == pulldom.CHARACTERS and \
                 node.nodeValue != '\n':
                    part.value += node.nodeValue


        #check the kpf version number
        if not 'kpf_version' in self._attributes:
            self._attributes['kpf_version'] = KPF_VERSION_STRING
            dirtyAtEnd = 1

        self._loaded_from_url = self._url
        self.set_url(url)

        for part in keybound_parts:
            self._getObserverSvc().notifyObservers(part,'kb-load', part.id)

        # this kicks off background status checking for the projects base path
        # the project url's will be added later, from partWrapper.js
        try:
            self._getObserverSvc().notifyObservers(self,'file_project',os.path.dirname(self._url))
        except Exception, unused:
            # if noone listens, exception
            pass


        self.set_isDirty(dirtyAtEnd)
        
        dt = time.clock() - t1
        log.info("project.load\t%4.4f" % (dt))

    def save(self):
        try:
            # cstringio would save a little time, but doesn't support unicode
            import StringIO
            stream = StringIO.StringIO()
            try:
                stream.write('<?xml version="1.0" encoding="UTF-8"?>'+newl)
                stream.write('<!-- Komodo Project File - DO NOT EDIT -->'+newl)
                self.serialize(stream)
            except Exception, e:
                log.exception(e)
                raise
            try:
                import tempfile
                # use a tempfile to avoid disasters if there's a failure on write.
                tempname = tempfile.mktemp()
                f = open(tempname, 'wb+')
                f.write(stream.getvalue().encode('UTF-8'))
                f.close()
            # if we get this far we have probably saved it correctly.
                fname = uriparse.URIToLocalPath(self._url)
                dirname = os.path.dirname(fname)
                if not os.path.exists(dirname):
                    os.makedirs(dirname)
                shutil.copyfile(tempname, fname)
                self.set_isDirty(0)
                self._loaded_from_url = self._url
            finally:
                os.unlink(tempname)
        except Exception, e:
            self.lastErrorSvc.setLastError(0, str(e))
            raise ServerException(nsError.NS_ERROR_ILLEGAL_VALUE, str(e))

    def revert(self):
        self._url = self._loaded_from_url
        self.load(self._url)
        self.set_isDirty(0)
        
    def dump(self, indent):
        print " "*indent + "project:"
        for x in self.__dict__.keys():
            print ' '*(indent+1) + x + ':' + str(self.__dict__[x])
        koFolderPart.dump(self, indent)

    def getAllContainedURLs(self):
        return self._urlmap.keys()

    def registerChildByURL(self, child):
        if child.hasAttribute('url'):
            url = child.getStringAttribute('url')
            self._urlmap[url] = child
        child = UnwrapObject(child) # this is necessary for the getattr to work.
        child.project = self
        kids = getattr(child, 'children', None)
        if kids:
            # You don't just adopt someone, you adopt their extended family
            for _child in kids:
                self.registerChildByURL(_child)
        # some children don't-- that's ok.

    def forgetChildByURL(self, child):
        if child.hasAttribute('url'):
            url = child.getStringAttribute('url')
            if url in self._urlmap:
                del self._urlmap[url]
        # forget decendants also
        kids = getattr(child, 'children', None)
        if kids:
            # You don't just adopt someone, you adopt their extended family
            for _child in kids:
                self.forgetChildByURL(_child)
                
    def getChildByURL(self, url):
        return self._urlmap.get(url, None)
    
    def close(self):
        toolbars = self.getChildrenByType('toolbar', 1)
        for toolbar in toolbars:
            self._getObserverSvc().notifyObservers(toolbar, 'toolbar_remove', 'toolbar_remove')
            
        menus = self.getChildrenByType('menu', 1)
        for menu in menus:
            self._getObserverSvc().notifyObservers(menu, 'menu_remove', 'menu_remove')
        self.set_isDirty(0)
        del self._urlmap
        self.destroy()

  
#---- Utility routines

# Code used to group filenames in language groups.
def _initKoLanguageRegistryService():
    global koLanguageRegistryService
    if koLanguageRegistryService is None:
        koLanguageRegistryService = components.classes["@activestate.com/koLanguageRegistryService;1"] \
            .getService(components.interfaces.koILanguageRegistryService);

def mapNamesToLanguages(names):
    if koLanguageRegistryService is None: _initKoLanguageRegistryService()
    newmap = {}
    for name in names:
        language = koLanguageRegistryService.suggestLanguageForFile(name)
        if not language: language = 'Other'
        newmap.setdefault(language, []).append(name)
    return newmap

class KoFileImportingService:
    _com_interfaces_ = [components.interfaces.koIFileImportingService]
    _reg_clsid_ = "{DC2BBD9F-A575-4ddc-8A22-D345B5413874}"
    _reg_contractid_ = "@activestate.com/koFileImportingService;1"
    _reg_desc_ = "Service to import files into folders"

    def __init__(self):
        self.lastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"]\
            .getService(components.interfaces.koILastErrorService)
        self.lastErrorSvc.setLastError(0, '')
        
    def findCandidateFiles(self, part, dirname, include, exclude, recursive):
        self.lastErrorSvc.setLastError(0, '')
        try:
            newfiles = []
            if include != '':
                include_pats = include.split(';')
            else:
                include_pats = None
            if exclude != '':
                exclude_pats = exclude.split(';')
            else:
                exclude_pats = None
            part = UnwrapObject(part)
            project = UnwrapObject(part._project)
            # start with current directory contents
            # clean up to deal with some samba problems
            # http://bugs.activestate.com/Komodo/show_bug.cgi?id=23097
            dirname = os.path.normpath(os.path.abspath(dirname))
            try:
                allnames = os.listdir(dirname)
            except OSError, e:
                if e.errno == 13: # permission denied
                    self.lastErrorSvc.setLastError(13, 'Permission denied')
                raise
                    
            allnames = _filterFiles(allnames, include_pats, exclude_pats)
            # filter out directories
            allnames = [os.path.join(dirname, name) for name in allnames\
                        if os.path.isfile(os.path.join(dirname, name))]
            # now look at children
            if recursive:
                for subdirname in _dirwalker(dirname, exclude_pats):
                    # clean up to deal with some samba problems
                    # http://bugs.activestate.com/Komodo/show_bug.cgi?id=23097
                    subdirname = os.path.normpath(os.path.abspath(subdirname))
                    try:
                        names = os.listdir(subdirname)
                    except OSError, e:
                        if e.errno == 13: # permission denied
                            self.lastErrorSvc.setLastError(13, 'Permission denied')
                            continue
                        raise
                    names = _filterFiles(names, include_pats, exclude_pats)
                    names = [os.path.join(subdirname, name) for name in names\
                                if os.path.isfile(os.path.join(subdirname, name))]
                    allnames.extend(names)
            for name in allnames:
                name = os.path.abspath(name)
                url = uriparse.localPathToURI(name)
                if not project._urlmap.get(url, None):
                    newfiles.append(name)
            return newfiles
        except:
            log.exception("failed scaning file system")

    def _getPart(self, filename, url, project):
        basename = os.path.basename(filename)
        if basename.endswith('.kpf'):
            return ProjectShortcut(url, basename, project)




        return File(url, basename, project)

    def addSelectedFiles(self, folder, importType, basedir, filenames):
        folder = UnwrapObject(folder)
        project = UnwrapObject(folder._project)
        _folderCache = {}
        filenames_and_urls = [(filename, uriparse.localPathToURI(str(filename))) for filename in filenames]
        data = [(self._getPart(filename, url, project), filename, url) for (filename,url) in filenames_and_urls]

        # Let the Code Intelligence system know that a file has been
        # added to one of Komodo's open projects or toolboxes.
        codeIntelSvc = components.classes["@activestate.com/koCodeIntelService;1"]\
            .getService(components.interfaces.koICodeIntelService);

        if importType == 'makeFlat':
            for part, filename, url in data:
                folder.children.append(part)
                project._urlmap[url] = part
                codeIntelSvc.ideEvent("added_file_to_project", url, None)
        elif importType == 'groupByType':
            registry = components.classes["@activestate.com/koLanguageRegistryService;1"].\
                getService(components.interfaces.koILanguageRegistryService)
            for part, filename, url in data:
                language = registry.suggestLanguageForFile(filename);
                if not language: language = 'Other'
                if language in _folderCache:
                    subfolder = _folderCache[language]
                else:
                    subfolder = _folderCache[language] = folder.getLanguageFolder(language)
                subfolder.children.append(part)
                project._urlmap[url] = part
                codeIntelSvc.ideEvent("added_file_to_project", url, None)
        elif importType == 'useFolders':
            for part, filename, url in data:
                subdirname = os.path.dirname(filename)
                if subdirname in _folderCache:
                    subfolder = _folderCache[subdirname]
                else:
                    subfolder = _folderCache[subdirname] = folder.getDirFolder(basedir, subdirname)
                subfolder.children.append(part)
                project._urlmap[url] = part
                codeIntelSvc.ideEvent("added_file_to_project", url, None)
        project.set_isDirty(len(data) > 0)

def _isOk(fullpath, exclude_pats):
    if exclude_pats:
        base = os.path.basename(fullpath)
        for pat in exclude_pats:
            if fnmatch.fnmatch(base, pat):
                return 0
    return 1

def _dirwalker(dirname, exclude_pats):
    """Generate all subdirectories of the given directory."""
    try:
        contents = os.listdir(os.path.normpath(dirname))
    except OSError, e:
        if e.errno != 13: # permission denied
            raise
        contents = []
    for f in contents:
        fullpath = os.path.join(dirname, f)
        if os.path.isdir(fullpath) and _isOk(fullpath, exclude_pats):
            yield fullpath
            for subdirname in _dirwalker(fullpath, exclude_pats):
                if _isOk(subdirname, exclude_pats):
                    yield subdirname

def _filterFiles(names, include_pats, exclude_pats):
    """Filter the names array for filenames which match any of the
    patterns in include_pats and match none of the patterns in
    exclude_pats -- returns a list containing a selection of the input
    names.
    
    if include_pats is None then all files match are considered by
    default
    
    if exclude_pats is None then no files are excluded explicitly
    """
    oknames = []
    for name in names:
        try:
            if include_pats is None:
                oknames.append(name)
                raise 'done'
            for pat in include_pats:
                if fnmatch.fnmatch(name, pat):
                    oknames.append(name)
                    raise 'done'
        except 'done':
            if exclude_pats:
                try:
                    for pat in exclude_pats:
                        if fnmatch.fnmatch(name, pat):
                            oknames.remove(name)
                            raise 'done'
                except 'done':
                    continue
    return oknames


_partFactoryMap = {}
for name, value in globals().items():
    if type(value) == types.ClassType and getattr(value, 'type', ''):
        _partFactoryMap[value.type] = value

def createPartFromType(type):
    return _partFactoryMap[type]()

def _getIntAttributeAndDelete(part, attribute_name):
    if attribute_name in part._attributes:
        number = int(part._attributes[attribute_name])
        del part._attributes[attribute_name]
    else:
        number = 0
    return number

def _SnippetConvert(part):
    # Old snippets (version 2.0 and 2.3b1) used attributes to refer to the
    # anchor and currentPos.  New snippets insert special markers in the
    # text itself, to make selection-preserving processing much easier
    # (unicode conversion, wrapping, etc.)
    # This code converts old snippet parts to the new ones and leaves the new ones
    # unchanged.
    if 'anchor_col' in part._attributes:
        anchor_line = _getIntAttributeAndDelete(part, 'anchor_line')
        anchor_col = _getIntAttributeAndDelete(part, 'anchor_col')
        currentPos_line = _getIntAttributeAndDelete(part, 'currentPos_line')
        currentPos_col = _getIntAttributeAndDelete(part, 'currentPos_col')
        _getIntAttributeAndDelete(part, 'anchor') # clean up of old junk
        _getIntAttributeAndDelete(part, 'currentPos') # clean up of old junk
        anchor = currentPos = 0
        lines = part.value.splitlines(1)
        # compute index-style anchor and currentPos from the (line,col) pairs
        # These coordinates are antiquated, and are removed from the part.
        for i in range(anchor_line):
            if len(lines) > i:
                anchor += len(lines[i])
        anchor += anchor_col
        for i in range(currentPos_line):
            if len(lines) > i:
                currentPos += len(lines[i])
        currentPos += currentPos_col
        if anchor < currentPos:
            part.value = part.value[:anchor] + \
                         ANCHOR_MARKER + \
                         part.value[anchor:currentPos] + \
                         CURRENTPOS_MARKER + \
                         part.value[currentPos:]
        else:
            part.value = part.value[:currentPos] + \
                         CURRENTPOS_MARKER + \
                         part.value[currentPos:anchor] + \
                         ANCHOR_MARKER + \
                         part.value[anchor:]
    return part


def evalPythonMacro(part, domdocument, window, scimoz, document, view, code):
    lastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"]\
        .getService(components.interfaces.koILastErrorService)
    lastErrorSvc.setLastError(0, '')
    import komodo
    komodo.domdocument = domdocument
    komodo.document = document
    komodo.editor = scimoz
    komodo.view = view
    komodo.window = window
    komodo.components = components
    komodo.macro = part
    macroGlobals = {}
    macroGlobals['komodo'] = komodo

    _partSvc = components.classes["@activestate.com/koPartService;1"]\
        .getService(components.interfaces.koIPartService)
    _partSvc.runningMacro = part

    lines = code.splitlines(1)
    lines = ['\t' + line + '\n' for line in lines]
    code = 'def _code():\n' + ''.join(lines) + '\n\n\n'
    try:
        exec code in macroGlobals, macroGlobals
        return eval('_code()', macroGlobals, macroGlobals)
    except Exception, e:
        lastErrorSvc.setLastError(1,
            ''.join(traceback.format_exception(*sys.exc_info())))
        raise ServerException


class KoPartService:
    _com_interfaces_ = [components.interfaces.koIPartService,
                        components.interfaces.nsIObserver]
    _reg_desc_ = "Komodo Part Service Component"
    _reg_contractid_ = "@activestate.com/koPartService;1"
    _reg_clsid_ = "{96DB159A-E772-4985-91B0-55A7FB7FEE19}"
    
    def __init__(self):
        self.lastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"].\
            getService(components.interfaces.koILastErrorService)
        self.wm = components.classes["@mozilla.org/appshell/window-mediator;1"].\
            getService(components.interfaces.nsIWindowMediator) 

        obsSvc = components.classes["@mozilla.org/observer-service;1"].\
                       getService(components.interfaces.nsIObserverService);
        obsSvc.addObserver(self, 'python_macro', 0);
        self.toolbox = None
        self.sharedToolbox = None
        self.runningMacro = None

    def getPartById(self, id):
        return _findPartById(id)

    def findPartForRunningMacro(self, partType, name, where='*'):
        """
        where: is one of
                "container"
                "toolbox"
                "shared toolbox"
                "toolboxes"
                "*"
            where "*" means "current part project, toolbox, shared toolbox"
        """
        
        if self.runningMacro:
            container = self.runningMacro.project
        else:
            container = None
        if where == '*':
            places = [container, self.toolbox, self.sharedToolbox]
        elif where == 'container':
            places = [container]
        elif where == 'toolbox':
            places = [self.toolbox]
        elif where == 'shared toolbox':
            places = [self.sharedToolbox]
        elif where == 'toolboxes':
            places = [self.toolbox, self.sharedToolbox]
        for place in places:
            if place:
                found = place.getChildWithTypeAndStringAttribute(
                            partType, 'name', name, 1)
                if found:
                    return found
        return None
    
    def observe(self, subject, topic, data):
        if topic != 'python_macro':
            return
        window = self.wm.getMostRecentWindow('Komodo');
        evalPythonMacro(None, None, window, None, data)

class koProjectPackageService:
    _com_interfaces_ = [components.interfaces.koIProjectPackageService]
    _reg_desc_ = "Komodo Packaging Service Component"
    _reg_contractid_ = "@activestate.com/koProjectPackageService;1"
    _reg_clsid_ = "{16237ba5-3d97-475b-ad22-fa60b2cd3a33}"
    
    def __init__(self):
        self.verbose = False
        self.percent = 10
        self.debug = 0
        self.test = 0
        self.lastErrorSvc = components.classes["@activestate.com/koLastErrorService;1"]\
            .getService(components.interfaces.koILastErrorService)
        self.fileSvc = components.classes["@activestate.com/koFileService;1"]\
            .getService(components.interfaces.koIFileService)

    def _gatherIcons(self, part):
        icons = []
        part = UnwrapObject(part)
        icon = part.get_iconurl()
        if self.test:
            print "icon [%s]"%icon
        if not icon.startswith('chrome://'):
            newicon = os.path.join('.icons', os.path.basename(icon))
            part.set_iconurl(newicon)
            icons.append((uriparse.URIToLocalPath(icon),newicon))
        if hasattr(part, 'getChildren'):
            for p in part.getChildren():
                icons += self._gatherIcons(p)
        return icons
    
    def packageProject(self, packagePath, project):
        if project.isDirty:
            err = 'Save project before packaging'
            self.lastErrorSvc.setLastError(1, err)
            raise ServerException(nsError.NS_ERROR_ILLEGAL_VALUE, err)

        self.packageParts(packagePath, project.getChildren())

    def _clonePartList(self, newproject, partList):
        # clone parts and add them to the project
        for part in partList:
            if part.type == 'project':
                self._clonePartList(newproject, part.getChildren())
            else:
                newpart = part.clone()
                newproject.addChild(newpart)
        
    # a list of parts instead of project
    def packageParts(self, packagePath, partList):
        import re
        # setup a temporary project file, as we may need to do modifications
        # that shoule only be in the packaged version
        if packagePath.find('.kpz') == -1:
            zipfilename = packagePath+'.kpz'
        else:
            zipfilename = packagePath
        if self.debug:
            print "zipfilename [%s]" % zipfilename
        if os.path.exists(zipfilename):
            err = 'A package with the same name already exists at that location.'
            self.lastErrorSvc.setLastError(1, err)
            raise ServerException(nsError.NS_ERROR_ILLEGAL_VALUE, err)

        try:
            projectName = os.path.splitext(os.path.basename(packagePath))[0]
            newproject = koProject()
            newproject.create()
            newproject._url = os.path.join(os.path.dirname(zipfilename), projectName+'.kpf')
            tmp_project_localPath = uriparse.URIToLocalPath(newproject._url)
            newproject.name = projectName
            if self.debug:
                print "newproject._url [%s]" % newproject._url

            # clone parts and add them to the project
            self._clonePartList(newproject, partList)
    
            # figure out a base path for all the parts
            newproject._relativeBasedir = os.path.commonprefix(newproject._urlmap.keys())
            if self.debug:
                print "using relative base [%s]" % newproject._relativeBasedir
    
            import zipfile
            if not self.test:
                zf = zipfile.ZipFile(str(zipfilename), 'w')
    
            # look at all the url's in the project, copy files, remove parts, etc.
            # as necessary
            fix_drive_re = re.compile(r'^(?:file:\/\/\/)?(?:(\w):)?(.*)$')
            for source in newproject._urlmap:
                part = UnwrapObject(newproject._urlmap[source])
                if 'url' in part._tmpAttributes and \
                   part._tmpAttributes['url'] == part._attributes['url']:
                    dest = part._tmpAttributes['relativeurl']
                else:
                    dest = uriparse.RelativizeURL(newproject._relativeBasedir, part._attributes['url'])
                diskfile = uriparse.URIToLocalPath(part._attributes['url'])
                if dest.find('file:')==0:
                    dest = fix_drive_re.sub(r'\1\2',dest)
                # we do not add directories
                if os.path.isfile(diskfile):
                    part._attributes['url'] = dest
                    dest = os.path.join(newproject.name, dest)
                    if self.debug:
                        print "diskfile [%r] dest[%r]" % (diskfile, dest)
                    if not self.test:
                        zf.write(str(diskfile), str(dest))
    
            # get a list of all the icons that are not in chrome so we can package
            # them
            iconlist = self._gatherIcons(newproject)
            for icondata in iconlist:
                source = icondata[0]
                dest = os.path.join(newproject.name, icondata[1])
                if self.debug:
                    print "icon diskfile [%r] dest[%r]" % (source,dest)
                if not self.test:
                    zf.write(str(source), str(dest))
            
            # save the temporary project file, add it to the zipfile, then delete it
            newproject.save()
            # save the new project
            project_filename = os.path.join(newproject.name, os.path.basename(tmp_project_localPath))
            if self.debug:
                print "writing project to zip as [%r]" % project_filename
            if not self.test:
                zf.write(str(tmp_project_localPath), str(project_filename))
        except Exception, e:
            if os.path.exists(tmp_project_localPath):
                os.unlink(tmp_project_localPath)
            err = 'An error occurred while attempting to export the package:\n\n%s' % e
            self.lastErrorSvc.setLastError(1, err)
            raise ServerException(nsError.NS_ERROR_ILLEGAL_VALUE, err)
            
        os.unlink(tmp_project_localPath)
        self.lastErrorSvc.setLastError(0, 'The package has been exported successfully to %s' % zipfilename)
    
    def importPackage(self, file, dir, part):
        projectName = os.path.splitext(os.path.basename(file))[0]
        projectFile = os.path.join(dir, projectName, projectName+'.kpf')
        packageDir = os.path.join(dir, projectName);
        
        if os.path.exists(packageDir) and not os.path.exists(projectFile):
            err = 'Unable to extract the package, the package directory [%s] already exists.' %packageDir
            self.lastErrorSvc.setLastError(1, err)
            raise ServerException(nsError.NS_ERROR_ILLEGAL_VALUE, err)
        
        try:
            if not os.path.exists(projectFile):
                self.extractPackage(file, dir)
            newproject = koProject()
            newproject.create()
            newproject.loadQuiet(projectFile)
            for child in newproject.getChildren():
                child = UnwrapObject(child)
                newchild = UnwrapObject(child.clone())
                if child._attributes.has_key('icon'):
                    # cloned parts have a relative url here, fix it
                    if newchild._attributes['icon'].find("://") == -1:
                        newchild._attributes['icon'] = uriparse.UnRelativizeURL(newproject._relativeBasedir, child._attributes['icon'])
                part.addChild(newchild)
        except Exception, e:
            err = 'An error occurred while attempting to import the package:\n\n%s' % e
            self.lastErrorSvc.setLastError(1, err)
            raise ServerException(nsError.NS_ERROR_ILLEGAL_VALUE, err)
        if not part.name:
            part.name = "Package %s" % projectName
        self.lastErrorSvc.setLastError(0, 'The package "%s" has been imported successfully.' % file)
        
    def extractPackage(self, file, dir):
        import zipfile
        if not dir.endswith(':') and not os.path.exists(dir):
            os.mkdir(dir)
        
        zf = zipfile.ZipFile(file)

        # create directory structure to house files
        self._createstructure(file, dir)

        num_files = len(zf.namelist())
        percent = self.percent
        divisions = 100 / percent
        perc = int(num_files / divisions)

        # extract files to directory structure
        i = 0
        for name in zf.namelist():
            i += 1
            #if self.verbose == True:
            #    print "Extracting %s" % name
            #elif (i % perc) == 0 and i > 0:
            #    complete = int (i / perc) * percent
            #    print "%s%% complete" % complete

            if not name.endswith('/'):
                outfile = open(os.path.join(dir, name), 'wb')
                outfile.write(zf.read(name))
                outfile.flush()
                outfile.close()


    def _createstructure(self, file, dir):
        self._makedirs(self._listdirs(file), dir)


    def _makedirs(self, directories, basedir):
        """ Create any directories that don't currently exist """
        for dir in directories:
            curdir = os.path.join(basedir, dir)
            if not os.path.exists(curdir):
                os.makedirs(curdir)

    def _listdirs(self, file):
        """ Grabs all the directories in the zip structure
        This is necessary to create the structure before trying
        to extract the file to it. """
        import zipfile
        zf = zipfile.ZipFile(file)

        dirs = {}

        for name in zf.namelist():
            if name.endswith('/'):
                dirs[name] = 1
            else:
                dirs[os.path.dirname(name)] = 1

        dirs = dirs.keys()
        dirs.sort()
        return dirs

