#!python
# Copyright (c) 2004 ActiveState Corp, a division of Sophos Plc.
# See the file LICENSE.txt for licensing information.

"""cb - utils for implementing a Code Browser UI with the 'codeintel' system

Using this module mainly involves working with a single CBRootNode
instance. You create one and then call its .{add|remove|update}Path()
methods which will maintain a structured AST representing all modules
defined in all added paths in the CIDB.
"""

import os
import sys
import logging
import types
import pprint
import bisect
import re

from codeintel.common import parseAttributes, symbolType2Name, canonicalizePath


#---- globals

log = logging.getLogger("cb")
#log.setLevel(logging.DEBUG)

__manager = None   # codeintel.Manager() instance, set by initialize()



#---- exceptions

class CBError(Exception):
    pass



#---- internal support stuff

class _DOMBuilder:
    def __init__(self, document):
        self._document = document

    def label(self, value, args={}, **kw):
        kw['value'] = value
        return self.element('label', args, **kw)
    def spring(self, args={}, **kw):
        return self.element('spring', args, **kw)
    def hbox(self, args={}, **kw):
        return self.element('hbox', args, **kw)

    def button(self, label, args={}, **kw):
        kw['label'] = label
        return self.element('button', args, **kw)

    def description(self, contents, args, **kw):
        elt = self.element('description', args, **kw)
        txt = self._document.createTextNode(contents)
        elt.appendChild(txt)
        return elt

    def element(self, tagname, args, **kw):
        _args = args.copy()
        _args.update(kw)
        b = self._document.createElement('label')
        b.setAttribute('flex', '0')
        for k,w in _args.items():
            b.setAttribute(k, str(w))
        #print "Creating label", value
        return b

    def add(self, parent, child):
        parent.appendChild(child)

def _fillDoc(docstring, dom, element):
    if docstring is None: docstring = ""
    data = []
    for line in docstring.splitlines():
        if not line.strip():
            if data:
                doc = dom.description('\r\n'.join(data),
                                      {'class': 'cb-doc-label'})
                dom.add(element, doc)
                data = []
        data.append(line)
    else:
        if data:
            doc = dom.description('\r\n'.join(data),
                                  {'class': 'cb-doc-label'})
            dom.add(element, doc)



#---- routine to determine images for Code Intel objects

def getImageURLForModule(language):
    images = {
        "Python" : "chrome://jaguar/skin/images/lang_python.png",
        "Perl": "chrome://jaguar/skin/images/lang_perl.png",
        "PHP": "chrome://jaguar/skin/images/lang_php.png",
        "Tcl": "chrome://jaguar/skin/images/lang_tcl.png",
        "JavaScript": "chrome://jaguar/content/icons/internet.png",
        "HTML": "chrome://jaguar/content/icons/insert_url.png",
        "XML": "chrome://jaguar/skin/images/lang_xml.png",
    }
    try:
        return images[language]
    except KeyError:
        return "chrome://jaguar/content/icons/draw-cylinder.png"

def getImageURLForSymbol(type, attributes):
    if type == "function":
        if 'private' in attributes:
            return "chrome://jaguar/skin/images/cb_function_private.png"
        if 'protected' in attributes:
            return "chrome://jaguar/skin/images/cb_function_protected.png"
        return "chrome://jaguar/skin/images/cb_function.png"
    elif type == "class":
        return getImageURLForClass(attributes)
    elif type == "variable":
        return getImageURLForVariable(attributes)
    elif type == "interface":
        if 'private' in attributes:
            return "chrome://jaguar/skin/images/cb_interface_private.png"
        if 'protected' in attributes:
            return "chrome://jaguar/skin/images/cb_interface_protected.png"
        return "chrome://jaguar/skin/images/cb_interface.png"
    else:
        raise CBError("unknown type name: %r" % type)

def getImageURLForClass(attributes):
    if "__namespace__" in attributes:
        return "chrome://jaguar/skin/images/cb_namespace.png"
    elif "private" in attributes:
        return "chrome://jaguar/skin/images/cb_class_private.png"
    elif "protected" in attributes:
        return "chrome://jaguar/skin/images/cb_class_protected.png"
    else:
        return "chrome://jaguar/skin/images/cb_class.png"

def getImageURLForVariable(attributes):
    if "__instancevar__" in attributes:
        if "private" in attributes:
            return "chrome://jaguar/skin/images/cb_instance_variable_private.png"
        if "protected" in attributes:
            return "chrome://jaguar/skin/images/cb_instance_variable_protected.png"
        return "chrome://jaguar/skin/images/cb_instance_variable.png"
    if "private" in attributes:
        return "chrome://jaguar/skin/images/cb_variable_private.png"
    if "protected" in attributes:
        return "chrome://jaguar/skin/images/cb_variable_protected.png"
    return "chrome://jaguar/skin/images/cb_variable.png"

def getImageURLForImport():
    return "chrome://jaguar/skin/images/cb_import.png"


def getDescForModule(name, path, language, attributes):
    if (language == "Python"
        and os.path.splitext(os.path.basename(path))[0] == "__init__"):
        thingy = "package"
    elif "__script__" in attributes:
        thingy = "script"
    elif language == "Tcl":
        thingy = "package"
    elif language == "Perl" and os.path.splitext(path)[1].lower() != ".pm":
        thingy = "script"
    else:
        thingy = "module"
    desc = "%s %s in '%s'" % (thingy, name, path)
    return desc

def getLangTypeNameForSymbol(typeName, attributes, language):
    """Convert a CodeIntel symbol name to a language-specific representative
    name. E.g., a CodeIntel "class" is used to represent a Perl "package".
    A CodeIntel "class" is sometimes -- depending on the attributes -- used
    to represent a Tcl "namespace".
    """
    langTypeName = typeName
    if typeName == "class":
        if "__namespace__" in attributes:
            langTypeName = "namespace"
        elif language == "Perl":
            langTypeName = "package"
    return langTypeName    

def getDescForSymbol(type, name, attributes, scope, language):
    if type == "class":
        desc = getLangTypeNameForSymbol(type, attributes, language)
        if "__namespace__" in attributes:
            desc = "namespace"
        elif language == "Perl":
            desc = "package"
        else:
            desc = type
        if scope:
            scopesep = {"Tcl": "::"}.get(language, ".")
            desc += ' ' + scopesep.join([scope, name])
        else:
            desc += ' '+name
        attributesForShow = [a for a in attributes if a != "__namespace__"]
        if attributesForShow:
            desc = ' '.join(attributesForShow) + ' ' + desc
    else:
        langTypeName = getLangTypeNameForSymbol(type, attributes, language)
        desc = "%s %s" % (langTypeName, name)
        if attributes:
            visible = [a for a in attributes.keys()
                       if not a.startswith("__") and not a.endswith("__")]
            if visible:
                desc = ' '.join(visible) + " " + desc
    return desc

def getDescForImport(module, symbol, alias, language):
    if language == "Tcl":
        #XXX Not handling symbol and alias
        s = "package require %s" % module
    elif language == "PHP":
        #XXX Not handling symbol and alias
        s = 'include "%s"' % module
    elif language == "Perl":
        if symbol == "*":
            s = "use %s" % module
        elif symbol == "**": # importing a tag, e.g. use Foo qw(:all);
            #XXX We *could* add a CIX <import> attribute at some point to
            #    specify the tag.
            s = "use %s qw(:<tag>)" % module
        elif symbol:
            s = "use %s qw(%s)" % (module, symbol)
        else:
            # This is either "use Foo ();" or "require Foo;". A search
            # the of the Perl 5.8 site lib should that the latter is about
            # 6 times more likely -- lets use that.
            #XXX We *could* add a CIX <import> attribute at some point to
            #    disambiguate.
            s = "require %s" % module
    else: # Just use Python naming for other languages
        if alias and symbol:
            s = "from %(module)s import %(symbol)s as %(alias)s" % locals()
        elif alias:
            s = "import %(module)s as %(alias)s" % locals()
        elif symbol:
            s = "from %(module)s import %(symbol)s" % locals()
        else:
            s = "import %(module)s" % locals()
    return s


#---- module initialization

def initialize(manager):
    global __manager
    __manager = manager

def _getManager():
    return __manager



#---- Code Browser (and Object Browser) tree "AST" nodes

class CBNode:
    def __init__(self):
        self.children = []
        # When sorting the tree alphabetically the "sortName" is used, if
        # defined, otherwise the name attribute is used. The line attribute
        # is used for sorting by file-order.
        self.name = None
        self.sortName = None
        self.line = None
        self._sortedBy = None

    #XXX This should change to be called "detail" to not conflict with the
    #    new name ("desc") of the one-liner descrption.
    def fillDescription(self, document, element):
        dom = _DOMBuilder(document)
        name = dom.label("Unknown", {'class': 'cb-signature-label'})
        dom.add(element, name)

    def pprint(self, level, indent):
        """Pretty print self and children to sys.stdout.

            "level" is the current indentation level
            "indent" is the indentation string to use (i.e.
                print indent*level)
        """
        raise NotImplementedError("virtual function: must override in subclass")

    def _sortNodes(self, nodes, sortBy):
        if self._sortedBy != sortBy:
            if sortBy == "file-order":
                nodes.sort(lambda a,b: cmp(a.line, b.line))
            elif sortBy == "name":
                nodes.sort(lambda a,b: cmp(a.sortName or a.name,
                                           b.sortName or b.name))
            else:
                log.warn("unknown sort by key (%r): ignoring", sortBy)
                return
            self._sortedBy = sortBy # cache sort order

    def generateRows(self, rows, nodeIsOpen, sortBy=None, level=0,
                     parentIndex=-1, hasNextSibling=0, scopedName=None,
                     filterString=None, nodeToLocate=None, cx=None, cu=None):
        """Traverse this node and its children (depth-first), appending
        rows as appropriate.

            "rows" is the list to which to add rows
            "nodeIsOpen" is a mapping of a node scoped-names to a boolean
                indicating if that node's children should be traversed. If
                there is no entry for a particular scoped-name then a
                reasonable default should be chosen.
            "sortBy" is a key describing how the row nodes should be sorted.
                Currently supported values are:
                    None: don't sort
                    "file-order": sort symbols by line
                    "name": sort symbols by name
                Implementations of this method in subclasses should just
                use the standard _sortNodes() method.
            "level" is the indentation level in the hierarchy
            "parentIndex" is the row number of the parent in the hierarchy
            "hasNextSibling" is a boolean indicating if that node has
                a sibling node for which a row will be subsequently added.
            "scopedName" is a mangling of the parent node data. This is used
                to handle persistence of twisty open/close state. Typically
                a child builds upon the scoped-name of its parent.
            "filterString" is a string specifying which nodes should be added
                or skipped.
            "nodeToLocate" is a 2-tuple indicating a node to find and show
                in the CB: (<'module' or 'symbol'>, <id>). Implementations
                must return the row number of the located row or None if the
                node was not located. Only one of "nodeToLocate" and
                "filterString" can be defined at once.
            "cx" and "cu", if specified, are a borrowed CIDB connection
                and cursor to use.

        Each 'row' is a dictionary with, at least, the following keys:
            {"node": <CBNode instance>,
             "codebrowser-tree-node": <node name string>,
             "level": <numeric level in tree hierarchy>,
             "parent-index": <row number of parent in hierarchy>}

        Any node that is a container (i.e. has children) must also defined
        "scoped-name" and "node-is-open". "scoped-name" is a fully scoped name
        of the node for unique identification. Subclasses should build upon
        their parent's "scopedName". "node-is-open" is a boolean indicating
        if the twisty is open and the children's rows are displayed.

        Any node for which "hasNextSibling" is true must added a
        "next-sibling-index" key which, obviously, indicates the row number
        of that next sibling.

        Any node may optionally include a "detail" key, which is a longer
        more descriptive string than the short "codebrowser-tree-node".
        """
        raise NotImplementedError("virtual function: must override in subclass")

    def matchesFilter(self, filterString):
        """Returns a boolean indicating if this node matches the filter."""
        return self.name.lower().find(filterString.lower()) != -1


class CBContainerNode(CBNode):
    """Base class for CBNodes that are just there to contain a groups of
    real code object elements.
    """
    type = langType = "container"
    imageURL = "chrome://jaguar/content/icons/open.png"
    title = ''
    file_path = ''

    def fillDescription(self, document, element):
        dom = _DOMBuilder(document)

        hbox = dom.hbox({'class': 'cb-scope-box'})
        if self.file_path:
            label = os.path.basename(self.file_path)
            fname = dom.label(label, {'class': 'cb-scope-label'})
            dom.add(hbox, fname)
        dom.add(element, hbox)

        if self.title:
            sig = dom.description(self.title+" for "+os.path.basename(self.file_path),
                                  {'class': 'cb-signature-label'})
        dom.add(element, sig)

class CBRootNode(CBNode):
    def __init__(self):
        CBNode.__init__(self)
        self.cpath2file = {}  # mapping of each working set canonical path to its CBFileNode
        self.modules = []    # list of modules (ordered by lang and name)
        self.cpath2moduleMap = {
            # <canonical path>: {<moduleName>: <CBModuleNode>}
        }

    def generateRows(self, rows, nodeIsOpen, sortBy=None, level=0,
                     parentIndex=-1, hasNextSibling=0, scopedName=None,
                     filterString=None, nodeToLocate=None, cx=None, cu=None):
        if filterString is not None and nodeToLocate is not None:
            raise CBError("error generating CB rows: cannot set both "
                          "filterString, %r, and nodeToLocate, %r."
                          % (filterString, nodeToLocate))
        retval = None
        for i in range(len(self.modules)):
            module = self.modules[i]
            locatedAt = module.generateRows(
                rows, nodeIsOpen, sortBy, level, parentIndex,
                (i+1 != len(self.modules)), scopedName,
                filterString=filterString, nodeToLocate=nodeToLocate,
                cx=cx, cu=cu)
            if locatedAt is not None:
                retval = locatedAt
        return retval

    def _createFileNode(self, cu, cpath, language):
        """Return a CBFileNode for the given path.

            "cu" is a CIDB database cursor to use. This is used to determine
                the associated CIDB file id, if any.
            "cpath" is the full (canonical) path to the file.
            "language" is the file's language.
        """
        cu.execute("SELECT id, path FROM file WHERE compare_path=? "
                   "AND language_id=(SELECT id FROM language WHERE name=?)",
                   (cpath, language))
        rows = tuple(cu)
        if not rows:
            node = CBFileNode(cpath)
        elif len(rows) == 1:
            file_id, file_path = rows[0]
            node = CBFileNode(cpath, file_id, file_path, language)
        else:
            raise CBError("unexpected database state: two rows in "
                          "the 'file' table with the same path, '%s', and "
                          "language, '%s': %s",
                          cpath, language, rows)
        return node

    def _createModuleNodes(self, cu, file):
        """Return the appropriate CBModuleNodes for the given flie.

            "cu" is a database cursor to use.
            "file" is a CBFileNode instance

        Typically there is just one module per file, some language allow
        the definition of symbols in multiple modules from one file.
        """
        if file.id is None:
            # No scan data for this file. Presumably it is coming. Add a
            # place-holder node for this file until results are available.
            modules = [ CBPseudoModuleNode(file.cpath, state="pending",
                                           file_path=file.path,
                                           file_language=file.language) ]
        else:
            # There can be zero or more scans of this file (there can be
            # more than one if there are more than one language engine
            # drivers for this file). We gather all data and merge the
            # information based on module name -- different generators
            # can define data about the same module.
            cu.execute("SELECT id, generator, md5, error FROM scan "
                       "WHERE file_id=%d" % file.id)
            scans = []
            scanErrors = {} # <generator name> -> <error string>
            moduleData = {}  # <module name> -> <module data>
            for scan_id, scan_generator, scan_md5, scan_error in tuple(cu):
                # If there is an md5 value for a scan, then there is
                # scan data.
                if scan_md5:
                    cu.execute("SELECT id, file_id, name, line, attributes, doc "
                               "FROM module WHERE file_id=? AND scan_id=?",
                               (file.id, scan_id))
                    for id, file_id, name, line, attrStr, doc in cu:
                        #XXX I suspect that is meant to be:
                        #       m = moduleData.setdefault(name, {})
                        #    so that there is one dict per different module
                        #    defined in this file.
                        m = moduleData.setdefault("name", {})
                        if "ids" not in m: m["ids"] = []
                        m["ids"].append(id)
                        m["file_id"] = file_id
                        m["name"] = name
                        m["line"] = line
                        m["attributes"] = parseAttributes(attrStr)
                        if "doc" not in m: m["doc"] = ""
                        if doc:
                            m["doc"] += doc
                        m["file_cpath"] = file.cpath
                        m["file_path"] = file.path
                        m["file_language"] = file.language
                if scan_error:
                    scanErrors[scan_generator] = scan_error

            # If there is scan
            # data from any of the generators then we create real modules
            # nodes accordingly. If any of the scans failed then each real
            # module nodes is given this data. If there are no real module
            # nodes then a pseudo-module node is created to make sure the
            # error is visible to the user.
            modules = []
            if not moduleData:
                m = CBPseudoModuleNode(file.cpath, state="error",
                                       scan_errors=scanErrors,
                                       file_id=file.id,
                                       file_path=file.path,
                                       file_language=file.language)
                modules.append(m)
            else:
                for name, data in moduleData.items():
                    m = CBModuleNode(scan_errors=scanErrors, **data)
                    modules.append(m)
        return modules

    def addPath(self, cu, path, language, lazy=True):
        """Add the given path to the CB tree.

            "cu" is an open CIDB cursor to use.
            "path" is the full path of the file to add to the CB tree
            "language" is the file's language.

        This adds all modules found in the CIDB for this path. A
        "pseudo-module" is added for this file if the CIDB does not have
        data for it (see CBPseudoModuleNode below).  Raises a CBError if
        there is a problem adding the file or modules for it (e.g.
        unexpected CIDB structure or content).
        """
        file = self._createFileNode(cu, path, language)
        modules = self._createModuleNodes(cu, file)

        self.cpath2file[file.cpath] = file
        moduleMap = {}  # module name -> module
        for module in modules:
            moduleMap[module.name] = module
            if not lazy:
                module.expand(cu)
            bisect.insort(self.modules, module)
        self.cpath2moduleMap[path] = moduleMap

    def removePath(self, cu, path, language=None):
        file = self.cpath2file[path]
        del self.cpath2file[path]
        moduleMap = self.cpath2moduleMap[path]
        del self.cpath2moduleMap[path]

        for name, module in moduleMap.items():
            self._removeModule(module, file.id)

    def addModule(self, cu, module, lazy=False):
        cpath = module.file_cpath
        if not lazy:
            module.linkUp(cu)
        bisect.insort(self.modules, module)
        if cpath not in self.cpath2moduleMap:
            self.cpath2moduleMap[cpath] = {}
        self.cpath2moduleMap[cpath][module.name] = module

    def _removeModule(self, module, file_id):
        """Remove the given module from self.modules."""
        index = bisect.bisect_left(self.modules, module)
        for i in range(index, len(self.modules)):
            if self.modules[i].file_id == file_id:
                del self.modules[i]
                break
        else:
            raise CBError("could not find module '%s' for file id %d in "
                          "CBRootNode module list" % (module.name, file_id))

    def updatePath(self, cu, path, language):
        try:
            oldFile = self.cpath2file[path]
        except KeyError:
            # This path has been removed in the interim (this can happen if
            # you switch to a file, initiating a re-scan, and then close
            # it quickly). Just drop the update.
            return
        oldModuleMap = self.cpath2moduleMap[path]

        # Update/add/remove the module nodes as required.
        #XXX This may be overkill. Unless the "update" path grows to include
        #    a calculation of specifically what symbols/rows have changed
        #    (if any) then just removing all modules and re-adding them
        #    all would be much more straight-forward.
        file = self._createFileNode(cu, path, language)
        modules = self._createModuleNodes(cu, file)
        self.cpath2file[file.cpath] = file
        moduleMap = {}  # module name -> module
        for module in modules:
            moduleMap[module.name] = module
            module.linkUp(cu)
            if module.name in oldModuleMap.keys():
                # Update the module
                #XXX This may eventually differ from the "Add the module"
                #    code path below to optimize which <tree/>
                #    rows need updating.
                self._removeModule(oldModuleMap[module.name], oldFile.id)
                bisect.insort(self.modules, module)
                del oldModuleMap[module.name]
            else: # Add the module
                bisect.insort(self.modules, module)
        for name, module in oldModuleMap.items():
            self._removeModule(module, oldFile.id) # Remove the module
        self.cpath2moduleMap[path] = moduleMap

    def __repr__(self):
        length = len(self.cpath2file.values())
        return "<CBRootNode: %d file%s>"\
               % (length, (length != 1 and "s" or ""))

    def pprint(self, level=0, indent="  "):
        print "%r" % self
        level += 1
        for module in self.modules:
            print "%s%r" % (indent*level, module)
            for symbol in module.children:
                symbol.pprint(level+1, indent)


class CBFileNode(CBNode):
    # Note: This node type never actually gets displayed in the Code Browser,
    #       so we don't have to bother with the displaying methods.
    type = langType = "file"
    def __init__(self, cpath, id=None, path=None, language=None):
        """Create a file node with the given path

            "cpath" is the full canonical path to the file
            "path" is the unmodified path (e.g. still has case on Windows)
            "id" is the CIDB row id for this file. If there is not CIDB
                entry for this file, then id==None.
            "language" is the content-type of the file.
        """
        CBNode.__init__(self)
        self.cpath = cpath
        self.path = path
        self.id = id
        self.language = language

    def __repr__(self):
        return "<CBFileNode '%s': id=%r language=%r>"\
               % (os.path.basename(self.cpath), self.id, self.language)


class CBModuleNode(CBNode):
    type = langType = "module"
    linkedUp = False # this is set to true when linkUp is called
    def __init__(self, ids, file_id, name, line, attributes, doc, file_cpath,
                 file_path, file_language, scan_errors):
        CBNode.__init__(self)
        self.ids = ids
        self.file_id = file_id
        self.name = name
        self.line = line
        self.attributes = attributes
        self.doc = doc
        self.file_cpath = file_cpath
        self.file_path = file_path
        self.file_language = file_language
        self.scan_errors = scan_errors
        self.nodes = None  # list of symbol (and import) nodes in this module
        self.imageURL = getImageURLForModule(file_language)
        self.desc = getDescForModule(name, file_path, file_language,
                                     attributes)

    def fillDescription(self, document, element):
        dom = _DOMBuilder(document)

        hbox = dom.hbox({'class': 'cb-scope-box'})
        if self.file_path:
            label = os.path.basename(self.file_path)
            fname = dom.label(label, {'class': 'cb-scope-label'})
            dom.add(hbox, fname)
        dom.add(element, hbox)

        sig = dom.description(self.name, {'class': 'cb-signature-label'})
        dom.add(element, sig)

        _fillDoc(self.doc, dom, element)

    def generateRows(self, rows, nodeIsOpen, sortBy=None, level=0,
                     parentIndex=-1, hasNextSibling=0, scopedName=None,
                     filterString=None, nodeToLocate=None, cx=None, cu=None):

        if (nodeToLocate or filterString) and not self.linkedUp:
            # we need the child nodes
            self.linkUp(cu)

        scopedName = "[%s]%s" % (self.file_cpath, self.name)
        row = {"codebrowser-tree-node": self.name,
               "detail": self.desc,
               "level": level,
               "parent-index": parentIndex,
               "scoped-name": scopedName,
               "node": self}
        #XXX Can't effectively include these until we have multi-line
        #    tooltips in the Code Browser <tree/>
        #if self.scan_errors:
        #    errmsgs = []
        #    for generator, error in self.scan_errors.items():
        #        errmsgs.append("error scanning '%s' with '%s' generator: %s"
        #                       % (self.name, generator, error))
        #    row["detail"] += '\n' + '\n'.join(errmsgs)
        rows.append(row)
        parentIndex = len(rows)-1

        if (nodeToLocate and nodeToLocate[0] == "module"
            and nodeToLocate[1] in self.ids):
            retval = len(rows)-1
            nodeToLocate = None
        else:
            retval = None

        if nodeToLocate:
            nBefore = len(rows)
            self._sortNodes(self.children, sortBy)
            for i in range(len(self.children)):
                child = self.children[i]
                locatedAt = child.generateRows(
                    rows, nodeIsOpen, sortBy, level+1, parentIndex,
                    (i+1 != len(self.children)), scopedName,
                    filterString=filterString, nodeToLocate=nodeToLocate,
                    cx=cx, cu=cu)
                if locatedAt:
                    row["node-is-open"] = nodeIsOpen[scopedName] = 1
                    retval = locatedAt
                    nodeToLocate = None
            if "node-is-open" not in row:
                row["node-is-open"] = nodeIsOpen.get(scopedName, 0)
            if not row["node-is-open"]:
                while len(rows) > nBefore:
                    rows.pop()
        elif filterString:
            row["node-is-open"] = 1
            self._sortNodes(self.children, sortBy)
            before = len(rows)
            for i in range(len(self.children)):
                child = self.children[i]
                child.generateRows(rows, nodeIsOpen, sortBy, level+1,
                                   parentIndex, (i+1 != len(self.children)),
                                   scopedName, filterString=filterString,
                                   cx=cx, cu=cu)
            after = len(rows)
            if before == after and not self.matchesFilter(filterString):
                rows.pop() # no matches in children
        else:
            row["node-is-open"] = nodeIsOpen.get(scopedName, 0)
            if row["node-is-open"]:
                if not self.linkedUp:
                    self.linkUp(cu)
                self._sortNodes(self.children, sortBy)
                for i in range(len(self.children)):
                    child = self.children[i]
                    child.generateRows(rows, nodeIsOpen, sortBy, level+1,
                                       parentIndex, (i+1 != len(self.children)),
                                       scopedName, filterString=filterString,
                                       cx=cx, cu=cu)
        if hasNextSibling:
            row["next-sibling-index"] = len(rows)
        return retval

    def __repr__(self):
        return "<CBModuleNode '%s': ids=%s file_id=%s>"\
               % (self.name, ','.join(map(str, self.ids)), self.file_id)

    def __cmp__(self, other):
        # Modules sort by module language, then by name. Pseudo-modules
        # generally do NOT know the file language: ensure those ones are last.
        if self.file_language is None and other.file_language is not None:
            return 1
        elif self.file_language is not None and other.file_language is None:
            return -1
        else:
            return cmp(self.file_language, other.file_language)\
                   or cmp(self.name.lower(), other.name.lower())

    def _getDisplayNodes(self, cu=None):
        """Build self.nodes from the CIDB.

            "cu" is a CIDB cursor to use. If None, then we get one from
                the configured Manager.
        """
        nodes = []

        if cu is None:
            borrowedCursor = False
            manager = _getManager()
            cx = manager.getDatabaseConnection()
            cu = cx.cursor()
            manager.acquireCIDBReadLock()
        else:
            borrowedCursor = True
        try:
            # Get a list of all variables. Afterwards, add a node only for those
            # types of scopes that we want to. (I.e. we exclude function
            # variables.)
            scopesThatShowVariables = {0: 1} # 0 == module scope id
            variables = []
            for id in self.ids:
                cu.execute("SELECT id, file_id, module_id, scope_id, scope, "
                           "       name, signature, line, type, attributes, doc "
                           "  FROM symbol "
                           " WHERE file_id=? AND module_id=?",
                           (self.file_id, id))
                for (id, file_id, module_id, scope_id, scope, name, signature,
                     line, type, attrStr, doc) in cu:
                    typeName = symbolType2Name(int(type))
                    scope_id = int(scope_id)
                    args = (int(id), int(file_id), int(module_id), scope_id,
                            scope, name, signature, line, typeName,
                            parseAttributes(attrStr), doc, self.file_cpath,
                            self.file_path, self.file_language)
                    if typeName in ("class", "interface"):
                        scopesThatShowVariables[id] = 1
                    classes = {"class": CBClassNode,
                               "function": CBFunctionNode,
                               "interface": CBInterfaceNode}
                    if typeName in classes:
                        node = classes[typeName](*args)
                        nodes.append(node)
                    elif typeName == "variable":
                        variables.append( (scope_id, args) )
            for scope_id, args in variables:
                if scope_id in scopesThatShowVariables:
                    node = CBVariableNode(*args)
                    nodes.append(node)
    
            for id in self.ids:
                cu.execute("SELECT id, file_id, module_id, scope_id, "
                           "       line, module, symbol, alias "
                           "  FROM import "
                           " WHERE file_id=? AND module_id=?",
                           (self.file_id, id))
                for (id, file_id, module_id, scope_id, line, module, name,
                     alias) in cu:
                    node = CBImportNode(int(id), int(file_id), int(module_id),
                                        int(scope_id), line, module, name, alias,
                                        self.file_cpath, self.file_path,
                                        self.file_language)
                    nodes.append(node)
        finally:
            if not borrowedCursor:
                manager.releaseCIDBReadLock()
                cu.finalize()
                cx.close()

        self.nodes = nodes
        if log.isEnabledFor(logging.DEBUG):
            log.debug("tree: symbols and imports for module '%s':\n%s",
                      self.name, pprint.pformat(nodes))


    def linkUp(self, cu=None):
        """Create a tree structure of the nodes based on their scope.

        This means setting up the .children (and .parent?) attributes of
        the symbols.
        """
        self._getDisplayNodes(cu)
        nodes = self.nodes
        nodes.sort(lambda a,b: cmp(a.scope_id, b.scope_id)) # sort so only need one pass
        scope_id = None
        parent = None
        for node in nodes:
            node.children = []
        scopeId2ImportsContainer = {}
        scopeId2InstanceVarsContainer = {}
        scopeId2GlobalVarsContainer = {}
        for node in nodes:
            if node.scope_id != scope_id:
                # This is a new scope, find the parent node.
                scope_id = node.scope_id
                if not scope_id: # top-level node
                    parent = self
                else:
                    #XXX Could use binary search here if this is too slow.
                    for p in nodes:
                        if p.id == scope_id:
                            parent = p
                            break
                    else:
                        raise CBError("error in nodes for module '%s': "
                                      "could not find parent for node: "
                                      "node=%r scope_id=%r nodes=\n%s"
                                      % (self.name, node, scope_id,
                                         pprint.pformat(nodes)))

            # Add this node to the parent's children.
            if isinstance(node, CBImportNode):
                # Put all imports into a separate "Imports" container.
                container = scopeId2ImportsContainer.get(scope_id)
                if not container:
                    container = CBImportsContainerNode(scope_id,
                                                       self.file_cpath,
                                                       self.file_path,
                                                       self.file_language)
                    parent.children.append(container)
                    scopeId2ImportsContainer[scope_id] = container
                container.children.append(node)
            elif isinstance(node, CBVariableNode) and node.isInstanceVar:
                # Put all instance vars into a separate container
                container = scopeId2InstanceVarsContainer.get(scope_id)
                if not container:
                    container = CBInstanceVarContainerNode(self.ids[0],
                                                           scope_id,
                                                           self.file_cpath,
                                                           self.file_path,
                                                           self.file_language)
                    parent.children.append(container)
                    scopeId2InstanceVarsContainer[scope_id] = container
                container.children.append(node)
            elif isinstance(node, CBVariableNode) and node.scope_id == 0:
                # Put all module-level variables in a "Global Vars" container
                container = scopeId2GlobalVarsContainer.get(scope_id)
                if not container:
                    container = CBGlobalVarContainerNode(self.ids[0],
                                                         scope_id,
                                                         self.file_cpath,
                                                         self.file_path,
                                                         self.file_language)
                    parent.children.append(container)
                    scopeId2GlobalVarsContainer[scope_id] = container
                container.children.append(node)
            else:
                parent.children.append(node)
        self.linkedUp = True


#XXX Should break this up into separate "CBScanErrorPseudoModuleNode" and
#    CBPendingScanPseudoModuleNode". THere is very little in common in the
#    two now.
class CBPseudoModuleNode(CBModuleNode):
    """A node at the module-level of the C.B. tree representing a file
    without scan data.

    A pseudo-module is one of two states: "pending" (scan data is
    anticipated) and "error" (previous scan(s) failed).
    """
    type = langType = "pseudo-module"
    def __init__(self, cpath, state="pending", scan_errors=None, file_id=None,
                 file_path=None, file_language=None):
        CBNode.__init__(self)
        self.cpath = cpath
        self.state = state
        self.scan_errors = scan_errors  # mapping of <generator name>: <error string>
        self.name = os.path.basename(file_path or cpath)
        self.file_id = file_id
        self.file_path = file_path
        self.file_language = file_language
        self.imageURL = self._getImageURL(self.state)

    def _getImageURL(self, state):
        if state == "pending":
            return "chrome://jaguar/skin/images/cb_scanning.png"
        elif state == "error":
            return "chrome://jaguar/skin/images/cb_error.png"
        else:
            raise ValueError("unexpected state for pseudo-module node: '%s'"
                             % state)

    def pprint(self, level=0, indent="  "):
        print level*indent + "Pseudo module: %s, %s" % (self.name, self.file_language)

    def generateRows(self, rows, nodeIsOpen, sortBy=None, level=0,
                     parentIndex=-1, hasNextSibling=0, scopedName=None,
                     filterString=None, nodeToLocate=None, cx=None, cu=None):
        if filterString and not self.matchesFilter(filterString):
            return

        row = {"codebrowser-tree-node": self.name,
               "level": level,
               "parent-index": parentIndex,
               "node": self}
        if self.state == "pending":
            row["detail"] = "scanning '%s'..." % self.name
        elif self.state == "error":
            errmsgs = []
            for generator, error in self.scan_errors.items():
                errmsgs.append("error scanning '%s' with '%s' generator: %s"
                               % (self.name, generator, error))
            row["detail"] = '\n'.join(errmsgs)
        rows.append(row)
        if hasNextSibling:
            row["next-sibling-index"] = len(rows)

    def __repr__(self):
        return "<CBPseudoModuleNode '%s': state='%s' file_id=%r>"\
               % (os.path.basename(self.file_path or self.cpath),
                  self.state, self.file_id)

    def fillDescription(self, document, element):
        dom = _DOMBuilder(document)

        hbox = dom.hbox({'class': 'cb-scope-box'})

        fname = dom.label(self.name, {'class': 'cb-scope-label'})
        dom.add(hbox, fname)
        dom.add(element, hbox)
        if self.state == "pending":
            sig = '(scanning...)'
        elif self.state == "error":
            sig = "error scanning '%s'" % self.name
        sigelt = dom.description(sig, {'class': 'cb-signature-label'})
        dom.add(element, sigelt)

        if self.state == "error":
            errmsgs = []
            for generator, error in self.scan_errors.items():
                _fillDoc("%s generator: %s" % (generator, error),
                         dom, element)

    def linkUp(self, cu=None):
        pass # don't have scan data, therefore no symbols to link-up


class CBVarContainerNode(CBContainerNode):
    imageURL = "chrome://jaguar/skin/images/cb_variable.png"
    title = "Variables"
    scopePart = "[variables]"

    def __init__(self, module_id, scope_id, file_cpath, file_path,
                 file_language):
        CBNode.__init__(self)
        self.module_id = module_id
        self.scope_id = scope_id
        self.file_cpath = file_cpath
        self.file_path = file_path
        self.file_language = file_language

        # Set line and name attributes s.t. this sorts before anything
        # except "Imports" always.
        self.line = -9
        self.name = -9

    def __repr__(self):
        klass = str(self.__class__).split('.')[-1]
        return "<%s: scope_id=%s>" % (klass, self.scope_id)

    def pprint(self, level, indent="  "):
        print "%s%r" % (indent*level, self)
        for import_ in self.children:
            import_.pprint(level+1, indent)

    def generateRows(self, rows, nodeIsOpen, sortBy=None, level=0,
                     parentIndex=-1, hasNextSibling=0, scopedName=None,
                     filterString=None, nodeToLocate=None, cx=None, cu=None):
        scopedName += "." + self.scopePart
        row = {"codebrowser-tree-node": self.title,
               "level": level,
               "parent-index": parentIndex,
               "scoped-name": scopedName,
               "node": self}
        rows.append(row)
        parentIndex = len(rows)-1

        retval = None
        if nodeToLocate:
            nBefore = len(rows)
            self._sortNodes(self.children, sortBy)
            for i in range(len(self.children)):
                child = self.children[i]
                locatedAt = child.generateRows(
                    rows, nodeIsOpen, sortBy, level+1, parentIndex,
                    (i+1 != len(self.children)), scopedName,
                    filterString=filterString, nodeToLocate=nodeToLocate,
                    cx=cx, cu=cu)
                if locatedAt:
                    row["node-is-open"] = nodeIsOpen[scopedName] = 1
                    retval = locatedAt
                    nodeToLocate = None
            if "node-is-open" not in row:
                row["node-is-open"] = nodeIsOpen.get(scopedName, 0)
            if not row["node-is-open"]:
                while len(rows) > nBefore:
                    rows.pop()
        elif filterString:
            row["node-is-open"] = 1
            before = len(rows)
            self._sortNodes(self.children, sortBy)
            for i in range(len(self.children)):
                child = self.children[i]
                child.generateRows(rows, nodeIsOpen, sortBy, level+1,
                                   parentIndex, (i+1 != len(self.children)),
                                   scopedName, filterString=filterString,
                                   cx=cx, cu=cu)
            after = len(rows)
            if before == after:
                rows.pop() # no matches in children
        else:
            row["node-is-open"] = nodeIsOpen.get(scopedName, 0)
            if row["node-is-open"]:
                self._sortNodes(self.children, sortBy)
                for i in range(len(self.children)):
                    child = self.children[i]
                    child.generateRows(rows, nodeIsOpen, sortBy, level+1,
                                       parentIndex, (i+1 != len(self.children)),
                                       scopedName, filterString=filterString,
                                       cx=cx, cu=cu)
        if hasNextSibling:
            row["next-sibling-index"] = len(rows)
        return retval

    def fillDescription(self, document, element):
        dom = _DOMBuilder(document)

        hbox = dom.hbox({'class': 'cb-scope-box'})
        if self.file_path:
            label = os.path.basename(self.file_path)
            fname = dom.label(label, {'class': 'cb-scope-label'})
            dom.add(hbox, fname)
        dom.add(element, hbox)

        if self.title:
            sig = dom.description(self.title + " for " + os.path.basename(self.file_path),
                                  {'class': 'cb-signature-label'})
        dom.add(element, sig)

        _fillDoc(', '.join([child.name for child in self.children]),
                dom, element)

class CBGlobalVarContainerNode(CBVarContainerNode):
    imageURL = "chrome://jaguar/skin/images/cb_variable.png"
    title = "Global Variables"
    scopePart = "[globalvars]"

    def __init__(self, module_id, scope_id, file_cpath, file_path,
                 file_language):
        CBVarContainerNode.__init__(self, module_id, scope_id, file_cpath,
                                    file_path, file_language)
        if file_language == "Perl":
            # bug 36070: Because we are ignoring some lexical scopes in Perl
            # and grouping some vars into "Global Variables" that really
            # aren't, lets not call them global
            self.title = "Variables"

class CBInstanceVarContainerNode(CBVarContainerNode):
    """A Code Browser node to contain all instance CBVariableNodes (as
    opposed to the typical class attribute nodes) for a scope.
    """
    imageURL = "chrome://jaguar/skin/images/cb_instance_variable.png"
    title = "Instance Variables"
    scopePart = "[instancevars]"


class CBSymbolNode(CBNode):
    def __init__(self, id, file_id, module_id, scope_id, scope, name,
                 signature, line, type, attributes, doc, file_cpath,
                 file_path, file_language, module_name=''):
        CBNode.__init__(self)
        self.id = id
        self.file_id = file_id
        self.module_id = module_id
        self.scope_id = scope_id
        self.scope = scope
        self.name = name
        self.signature = signature
        self.line = line
        # Note that this "type" field is a type _name string_ rather than
        # one of the symbol type ST_* constants used for symbol types in
        # the CIDB.
        self.type = type
        self.attributes = attributes
        self.doc = doc
        self.file_cpath = file_cpath
        self.file_path = file_path
        self.file_language = file_language
        self.langType = getLangTypeNameForSymbol(type, attributes, file_language)
        self.imageURL = getImageURLForSymbol(type, attributes)
        self.desc = signature or getDescForSymbol(type, name, attributes,
                                                  scope, file_language)
        self.module_name = module_name

    def fillDescription(self, document, element):
        dom = _DOMBuilder(document)

        hbox = dom.hbox({'class': 'cb-scope-box'})

        if self.file_path:
            label = os.path.basename(self.file_path)
            if self.line:
                label = label + ':' + str(self.line)
            fname = dom.label(label, {'class': 'cb-scope-label'})
            dom.add(hbox, fname)
        dom.add(element, hbox)

        if self.signature:
            sig = dom.description(self.signature, {'class': 'cb-signature-label'})
            dom.add(element, sig)
        else:
            name = dom.description(self.name, {'class': 'cb-signature-label'})
            dom.add(element, name)

        _fillDoc(self.doc, dom, element)

    def generateRows(self, rows, nodeIsOpen, sortBy=None, level=0,
                     parentIndex=-1, hasNextSibling=0, scopedName=None,
                     filterString=None, nodeToLocate=None, cx=None, cu=None):
        scopedName += "."+self.name
        row = {"codebrowser-tree-node": self.name,
               "level": level,
               "parent-index": parentIndex,
               "scoped-name": scopedName,
               "detail": self.desc,
               "node": self}
        rows.append(row)
        parentIndex = len(rows) - 1

        if nodeToLocate and nodeToLocate[1] == self.id:
            retval = len(rows)-1
            nodeToLocate = None
        else:
            retval = None

        if nodeToLocate:
            nBefore = len(rows)
            self._sortNodes(self.children, sortBy)
            for i in range(len(self.children)):
                child = self.children[i]
                locatedAt = child.generateRows(
                    rows, nodeIsOpen, sortBy, level+1, parentIndex,
                    (i+1 != len(self.children)), scopedName,
                    filterString=filterString, nodeToLocate=nodeToLocate,
                    cx=cx, cu=cu)
                if locatedAt:
                    row["node-is-open"] = nodeIsOpen[scopedName] = 1
                    retval = locatedAt
                    nodeToLocate = None
            if "node-is-open" not in row:
                row["node-is-open"] = nodeIsOpen.get(scopedName, 0)
            if not row["node-is-open"]:
                while len(rows) > nBefore:
                    rows.pop()
        elif filterString:
            row["node-is-open"] = 1
            before = len(rows)
            self._sortNodes(self.children, sortBy)
            for i in range(len(self.children)):
                symbol = self.children[i]
                symbol.generateRows(rows, nodeIsOpen, sortBy, level+1,
                                    parentIndex, (i+1 != len(self.children)),
                                    scopedName, filterString=filterString,
                                    cx=cx, cu=cu)
            after = len(rows)
            match = self.name.lower().find(filterString.lower()) != -1
            if before == after and not match:
                rows.pop() # no match in self or children
        else: # just drill through the open ones.
            row["node-is-open"] = nodeIsOpen.get(scopedName, 0)
            if row["node-is-open"]:
                self._sortNodes(self.children, sortBy)
                for i in range(len(self.children)):
                    symbol = self.children[i]
                    symbol.generateRows(rows, nodeIsOpen, sortBy, level+1,
                                        parentIndex, (i+1 != len(self.children)),
                                        scopedName, filterString=filterString,
                                        cx=cx, cu=cu)
        if hasNextSibling:
            row["next-sibling-index"] = len(rows)
        return retval

    def __repr__(self):
        klass = str(self.__class__).split('.')[-1]
        return "<%s '%s': id=%s module_id=%s line=%s scope='%s'>"\
               % (klass, self.name, self.id, self.module_id,
                  self.line, self.scope)

    def pprint(self, level, indent="  "):
        print "%s%r" % (indent*level, self)
        for symbol in self.children:
            symbol.pprint(level+1, indent)


class CBClassNode(CBSymbolNode):
    pass

class CBFunctionNode(CBSymbolNode):
    pass

class CBInterfaceNode(CBSymbolNode):
    pass

class CBVariableNode(CBSymbolNode):
    def __init__(self, *args, **kwargs):
        CBSymbolNode.__init__(self, *args, **kwargs)
        # Handle special attributes for variables.
        self.imageURL = getImageURLForVariable(self.attributes)
        if "__instancevar__" in self.attributes:
            del self.attributes["__instancevar__"]
            self.isInstanceVar = 1
        else:
            self.isInstanceVar = 0


class CBImportsContainerNode(CBContainerNode):
    """A Code Browser node to contain all CBImportNodes for a scope."""
    imageURL = "chrome://jaguar/skin/images/cb_import.png"

    def __init__(self, scope_id, file_cpath, file_path, file_language):
        CBNode.__init__(self)
        self.scope_id = scope_id
        self.file_cpath = file_cpath
        self.file_path = file_path
        self.file_language = file_language

        if file_language == "Perl":
            self.title = "Uses and Requires"
        elif file_language == "PHP":
            self.title = "Includes and Requires"
        else:
            self.title = "Imports"

        # Set line and name attributes s.t. this sorts first always.
        self.line = -10
        self.name = -10

    def __repr__(self):
        return "<CBImportsContainerNode: scope_id=%s>" % self.scope_id

    def pprint(self, level, indent="  "):
        print "%s%r" % (indent*level, self)
        for import_ in self.children:
            import_.pprint(level+1, indent)

    def generateRows(self, rows, nodeIsOpen, sortBy=None, level=0,
                     parentIndex=-1, hasNextSibling=0, scopedName=None,
                     filterString=None, nodeToLocate=None, cx=None, cu=None):
        scopedName += ".[imports]"
        row = {"codebrowser-tree-node": self.title,
               "level": level,
               "parent-index": parentIndex,
               "scoped-name": scopedName,
               "node": self}
        rows.append(row)
        parentIndex = len(rows)-1
        if filterString:
            row["node-is-open"] = 1
            before = len(rows)
            self._sortNodes(self.children, sortBy)
            for i in range(len(self.children)):
                child = self.children[i]
                child.generateRows(rows, nodeIsOpen, sortBy, level+1,
                                   parentIndex, (i+1 != len(self.children)),
                                   scopedName, filterString=filterString,
                                   cx=cx, cu=cu)
            after = len(rows)
            if before == after:
                rows.pop() # no matches in children
        else:
            row["node-is-open"] = nodeIsOpen.get(scopedName, 0)
            if row["node-is-open"]:
                self._sortNodes(self.children, sortBy)
                for i in range(len(self.children)):
                    child = self.children[i]
                    child.generateRows(rows, nodeIsOpen, sortBy, level+1,
                                       parentIndex, (i+1 != len(self.children)),
                                       scopedName, filterString=filterString,
                                       cx=cx, cu=cu)
        if hasNextSibling:
            row["next-sibling-index"] = len(rows)

    def fillDescription(self, document, element):
        dom = _DOMBuilder(document)

        hbox = dom.hbox({'class': 'cb-scope-box'})
        if self.file_path:
            label = os.path.basename(self.file_path)
            fname = dom.label(label, {'class': 'cb-scope-label'})
            dom.add(hbox, fname)
        dom.add(element, hbox)

        if self.title:
            sig = dom.description(self.title + " for " + os.path.basename(self.file_path),
                                  {'class': 'cb-signature-label'})
        dom.add(element, sig)

        for child in self.children:
            dom.add(element, dom.description(child.desc,
                       {'class': 'cb-doc-contentline'}))

class CBImportNode(CBNode):
    type = langType = "import"

    def __init__(self, id, file_id, module_id, scope_id, line, module,
                 symbol, alias, file_cpath, file_path, file_language,
                 module_name=''):
        CBNode.__init__(self)
        self.id = id
        self.file_id = file_id
        self.module_id = module_id
        self.scope_id = scope_id
        self.line = line
        self.module = module
        self.symbol = symbol
        self.alias = alias
        self.file_cpath = file_cpath
        self.file_path = file_path
        self.file_language = file_language
        if alias:
            if symbol:
                name = "%s (%s.%s)" % (alias, module, symbol)
            else:
                name = "%s (%s)" % (alias, module)
        elif symbol:
            if symbol in ("*", "**") and file_language != "Python":
                # "*" and "**" are used as special indicator for other
                # languages (e.g. Perl) and aren't appropriate
                # representations for the user.
                name = module
            else:
                separator = {"Python": ".",
                             "Perl": "::",
                             "PHP": "::"}
                name = separator.get(file_language, ".").join([module, symbol])
        else:
            name = module
        self.name = name
        self.module_name = module_name
        self.imageURL = getImageURLForImport()
        self.desc = getDescForImport(module, symbol, alias, file_language)

    def __repr__(self):
        s = "<CBImportNode: id=%s scope_id=%s module='%s'"\
            % (self.id, self.scope_id, self.module)
        if self.symbol: s += " symbol='%s'" % self.symbol
        if self.alias: s += " alias='%s'" % self.alias
        if self.line: s += " line=%s" % self.line
        s += ">"
        return s

    def fillDescription(self, document, element):
        dom = _DOMBuilder(document)

        hbox = dom.hbox({'class': 'cb-scope-box'})

        if self.file_path:
            label = os.path.basename(self.file_path)
            if self.line:
                label = label + ':' + str(self.line)
            fname = dom.label(label, {'class': 'cb-scope-label'})
            dom.add(hbox, fname)
        dom.add(element, hbox)

        sig = dom.description(self.name,
                              {'class': 'cb-signature-label'})
        dom.add(element, sig)
        _fillDoc(self.desc, dom, element)

    def pprint(self, level, indent="  "):
        print "%s%r" % (indent*level, self)

    def generateRows(self, rows, nodeIsOpen, sortBy=None, level=0,
                     parentIndex=-1, hasNextSibling=0, scopedName=None,
                     filterString=None, nodeToLocate=None, cx=None, cu=None):
        if filterString and not self.matchesFilter(filterString):
            return

        scopedName += "."+self.module
        if self.symbol: scopedName += "."+self.symbol
        if self.alias: scopedName += "."+self.alias
        row = {"codebrowser-tree-node": self.name,
               "level": level,
               "parent-index": parentIndex,
               "scoped-name": scopedName,
               "detail": self.desc,
               "node": self}
        rows.append(row)
        if hasNextSibling:
            row["next-sibling-index"] = len(rows)

