#!python
# Copyright (c) 2004 ActiveState Corp, a division of Sophos Plc.
# See the file LICENSE.txt for licensing information.

"""Code Intelligence: database handling

Usage:
    XXX how to use the CIDB
"""

import os, sys
import threading
import time
import xml.sax
from xml.sax.handler import ContentHandler, ErrorHandler
import md5
import pprint
import socket
import shutil

import logging
import process
import apsw

from codeintel.common import *
from codeintel import ScanRequest
from codeintel.util import parsePyFuncDoc, parseDocSummary


#---- globals

log = logging.getLogger("codeintel.database")
#log.setLevel(logging.DEBUG)



#---- internal support

def _version2versionInfo(version):
    verInfo = []
    for part in version.split('.'):
        try:
            part = int(part)
        except ValueError:
            pass
        verInfo.append(part)
    return tuple(verInfo)


_gLineSplitPat = re.compile(r"\r\n|\r|\n")
_gEncodingAttrPat = re.compile('encoding="([^"]*)"')
def _detectXMLEncoding(content):
    "Return the encoding of the given XML content or None if it cannot tell."
    if not content.startswith("<?xml"):
        return None
    import re
    firstline = _gLineSplitPat.split(content, 1)[0]
    match = _gEncodingAttrPat.search(firstline)
    if not match:
        return None
    encoding = match.group(1)
    return encoding


class _Cix2CidbHandler(ContentHandler):
    """SAX Event Content Handler to update a CIDB database with CIX."""
    MIN_CIX_VERSION = (0,1)

    def __init__(self, cx, cu, manager, scan_imports=True):
        """Create a CIX->CIDB importer.
        
            "cx" is the CIDB connection
            "cu" is the CIDB cursor to use
            "manager" is a the controlling codeintel.Manager.
            "scan_imports" (optional, default true) is a boolean indicating
                if the processing of file imports should be scheduled.
        """
        ContentHandler.__init__(self)
        self.cx = cx
        self.cu = cu
        self.manager = manager
        self.scan_imports = scan_imports
        # State variables
        self.tagstack = []          # stack of CIX tags
        self.scope_id_stack = None  # stack of symbol scope ids
        self.scope_stack = None     # stack of symbol scope names
        self.file_id = None         # CIDB row id for the current file
        self.module_id = None       # CIDB row id for the current module
        # used to accumulate character content in <doc> and <signature>
        self.doc = self.signature = None
        # The (<table>, <id>) of the item (a 'symbol' entry or a 'base'
        # entry) with which <type/> tag information can be associated.
        self.last_typeable_row = (None, None)
        
        # Remember data about the "last successful <file/>" for error
        # recovery.
        self.lsf_path = None
        self.lsf_cpath = None
        self.lsf_generator = None
        self.lsf_language = None
        
        # Info to use for invalidating caches on COMMIT.
        self.updated_files = [] 
        self.updated_perl_modules = []

    def startElement(self, tag, attrs):
        #if log.isEnabledFor(logging.DEBUG):
        #    attrDict = {}
        #    for key, value in attrs.items():
        #        attrDict[key] = value
        #    log.debug("cix: startElement %s %s", tag, attrDict)
        self.tagstack.append(tag)
        funcName = "start_"+tag
        func = getattr(self, funcName, None)
        if func is not None:
            func(attrs)
    def endElement(self, tag):
        #log.debug("cix: endElement %s", tag)
        funcName = "end_"+tag
        func = getattr(self, funcName, None)
        if func is not None:
            func()
        self.tagstack.pop()

    def start_codeintel(self, attrs):
        # validate the CIX version
        if not attrs.has_key("version"):
            raise CIXError("no 'version' tag on 'codeintel' element")
        version = attrs.get("version")
        versionInfo = _version2versionInfo(version)
        if versionInfo < self.MIN_CIX_VERSION:
            raise CIXError("CIX content version, %s, is unsupported" % version)
        #log.debug("cix: version %s", version)

    def start_file(self, attrs):
        """<file .../>
        
        Attributes:
            "path" (required) path of the file scanned. This path should
                not be made absolute if the given path was relative. The
                path must use Unix-style dir separators, '/', so that
                CIX can be cross-platform.
            "generator" (required) is the name of the language engine. This
                must match the name of the language engine driver (as
                defined in engines.py) exactly.
            "langauge" (required) is the language of the scanned content.
            "mtime" (required for successful scan) is a last modified time of
                the file in seconds since the "epoch".
            "md5" (required for successful scan) is the MD5 hexdigest of the
                scanned content.
            "error" (required for failed scan) is a string describing why the
                scan failed. This is intended to be an internal-only form of
                CIX.
        """
        # Must be child of <codeintel/> tag.
        if self.tagstack[-2] != "codeintel":
            raise CIXError("<file> tag must be an immediate child of a "
                           "<codeintel> tag")
        try:
            path = attrs["path"]
        except KeyError, ex:
            raise CIXError("<file> tag has no 'path' attribute")
        cpath = canonicalizePath(path)
        try:
            generator = attrs["generator"]
        except KeyError, ex:
            raise CIXError("<file> tag has no 'generator' attribute")
        try:
            language = attrs["language"]
        except KeyError, ex:
            raise CIXError("<file> tag has no 'language' attribute")
        if attrs.has_key("mtime"):
            mtime = attrs["mtime"]
        else:
            mtime = None
        if attrs.has_key("md5"):
            md5sum = attrs["md5"]
        else:
            md5sum = None
        if attrs.has_key("error"):
            error = attrs["error"]
        else:
            error = None
        if not ((mtime and md5sum and error is None) or
                (mtime is None and md5sum is None and error is not None)):
            raise CIXError("illegal <file> tag: must define 'mtime' and "
                           "'md5' for a successful scan or 'error' for a "
                           "failed scan")

        # Save these for error recovery.
        self.lsf_path = path
        self.lsf_cpath = cpath
        self.lsf_generator = generator
        self.lsf_language = language
        
        #log.debug("cix: update files and scans: (%r, %r, %r, %r, %r, %r)",
        #          cpath, generator, language, mtime, md5sum, error)
        cu = self.cu
        cu.execute("SELECT id FROM language WHERE name=? LIMIT 1", (language,))
        try:
            language_id = int(cu.next()[0])
        except StopIteration:
            # there is no row for this language, add one
            cu.execute("INSERT INTO language (name) VALUES (?)", (language,))
            language_id = self.cx.last_insert_rowid()
        self.language_id = language_id
        
        cu.execute("SELECT id FROM file WHERE compare_path=? "
                   "AND language_id=? LIMIT 1", (cpath, language_id))
        try:
            file_id = int(cu.next()[0])
        except StopIteration:
            # there is no row for this language, add one
            cu.execute("INSERT INTO file (path, compare_path, language_id) "
                       "VALUES (?, ?, ?)",
                       (canonicalizePath(path, False), cpath, language_id))
            file_id = self.cx.last_insert_rowid()
        self.file_id = file_id
            
        cu.execute("SELECT id FROM scan WHERE file_id=? AND generator=? LIMIT 1",
                   (file_id, generator))
        try:
            # update this scan row
            scan_id = int(cu.next()[0])
            if md5sum:
                cu.execute("UPDATE scan SET mtime=?, md5=?, error=? "
                           "WHERE id=?", (mtime, md5sum, error, scan_id))
            else:
                # Intentionally do NOT null out 'md5' value.
                cu.execute("UPDATE scan SET error=? WHERE id=?",
                           (error, scan_id))
        except StopIteration:
            # there is no scan entry for the file/generator combo, add one
            cu.execute("INSERT INTO scan (file_id, generator, mtime, md5, error) "
                       "VALUES (?, ?, ?, ?, ?)",
                       (file_id, generator, mtime, md5sum, error))
            scan_id = self.cx.last_insert_rowid()
        self.scan_id = scan_id

        self._old_imports = self._getFileImportData(cu, scan_id)
        # If the scan succeeded then, clear all previous scan data for this
        # file and generator in preparation for adding the data from
        # this CIX document.
        if md5sum:
            #starttime = time.clock()
            cu.execute("DELETE FROM module WHERE scan_id=%d" % scan_id)
            cu.execute("DELETE FROM symbol WHERE scan_id=%d" % scan_id)
            cu.execute("DELETE FROM base WHERE scan_id=%d" % scan_id)
            cu.execute("DELETE FROM import WHERE scan_id=%d" % scan_id)
            cu.execute("DELETE FROM type WHERE scan_id=%d" % scan_id)
            #endtime = time.clock()
            #print "XXX DELETEing old scan took %.3f sec" % (endtime-starttime)
    def end_file(self):
        # Schedule changed import statements for scanning.
        if self.scan_imports and self.lsf_language in self.manager.coreLanguages:
            new = self._getFileImportData(self.cu, self.scan_id)
            old = self._old_imports
            changed = [i for i in new if i not in old]
            if changed:
                handler = self.manager.getImportHandler(self.lsf_language)
                cwd = os.path.dirname(os.path.abspath(self.lsf_path))
                for (module, symbol) in changed:
                    mpath, kind, scannable = handler.findModuleOnDisk(
                        module, symbol, cwd=cwd)
                    if mpath and scannable:
                        log.info("request scan for import '%s'", mpath)
                        r = ScanRequest(mpath, self.lsf_language,
                                        PRIORITY_BACKGROUND)
                        self.manager.addRequest(r)
                        
        self.updated_files.append((self.file_id, self.lsf_cpath))
        self.file_id = None
        self.scan_id = None

    def _getFileImportData(self, cu, scan_id):
        """Return a unique list of import data for the given scan.
        Here "unique" means only the import row data relevant to identifying
        the file imported.
        """
        cu.execute("SELECT module, symbol FROM import WHERE scan_id=%d" % scan_id)
        unique = {}
        for row in cu:
            unique[tuple(row)] = 1
        #return unique.keys()
        return unique

    # Handle possible <doc/> and <signature/> children of any module or
    # symbol tag (e.g. <module>, <class>, <function>,...).
    def characters(self, content):
        # Note: I am specifically NOT unencoding the lower-ASCII char
        #       entities, for example: '&#8;' is not being changed
        #       to '\b'.
        if self.tagstack[-1] == "doc":
            #log.debug("cix: doc characters %r", content)
            if self.doc:
                self.doc += content
            else:
                self.doc = content
        elif self.tagstack[-1] == "signature":
            #log.debug("cix: signature characters %r", content)
            if self.signature:
                self.signature += content
            else:
                self.signature = content
    def end_doc(self):
        if self.doc:
            if self.tagstack[-2] == "module":
                table = "module"
                id_ = self.module_id
            else:
                table = "symbol"
                id_ = self.scope_id_stack[-1]
            doc = self.doc
            self.cu.execute("UPDATE %s SET doc=? WHERE id=?" % table,
                            (doc, id_))
        self.doc = None
    def end_signature(self):
        if self.signature:
            if self.tagstack[-2] == "module":
                table = "module"
                id_ = self.module_id
            else:
                table = "symbol"
                id_ = self.scope_id_stack[-1]
            signature = self.signature
            self.cu.execute("UPDATE %s SET signature=? WHERE id=?" % table,
                            (signature, id_))
        self.signature = None

    def start_module(self, attrs):
        """<module .../>
        
        Attributes:
            "name" (required)
            "line" (optional, default 1) is a 1-based line number on which
                the module definition begins
            "attributes" (optional, default null) is a space-separated list
                of attributes modifying this module.
        """
        if self.tagstack[-2] != "file":
            raise CIXError("<module> tag must be an immediate child of a "
                           "<file> tag")
        file_id = self.file_id
        scan_id = self.scan_id
        try:
            name = attrs["name"]
        except KeyError, ex:
            raise CIXError("<module> tag has no 'name' attribute")
        if attrs.has_key("line"):
            try:
                line = int(attrs["line"])
            except ValueError, ex:
                raise CIXError("illegal <module> 'line' attribute: must "
                               "be integer: %r" % attrs["line"])
        else:
            line = 1
        if attrs.has_key("attributes"):
            attributes = attrs["attributes"]
        else:
            attributes = None
        doc = None
        self.cu.execute("""INSERT INTO module
                            (file_id, scan_id, name, line, attributes, doc)
                           VALUES (:file_id, :scan_id, :name,
                                   :line, :attributes, :doc)""",
                        locals())
        if self.lsf_language == "Perl":
            self.updated_perl_modules.append(name)
        self.module_id = self.cx.last_insert_rowid()
        self.scope_id_stack = [None]
        self.scope_stack = []
    def end_module(self):
        self.scope_id_stack = None
        self.scope_stack = None
        self.module_id = None

    def start_function(self, attrs):
        """<function .../>
        
        Attributes:
            "name" (required)
            "line" (optional) is a 1-based line number on which the
                definition begins. This is only really optional to support
                CIX data for binary files where a line number doesn't
                make sense. For scanned files a line should really be
                specified.
            "lineend" (optional, though strongly recommended) is a 1-based
                line number for the last line on which the definition occurs.
            "attributes" (optional, default null) is a space-separated list
                of attributes modifying this function -- for example,
                "virtual public".
            "signature" (optional, string)
        """
        file_id = self.file_id
        scan_id = self.scan_id
        module_id = self.module_id
        scope_id = self.scope_id_stack[-1] or 0
        scope = '.'.join(self.scope_stack)
        try:
            name = attrs["name"]
        except KeyError, ex:
            raise CIXError("<function> tag has no 'name' attribute")
        if attrs.has_key("line"):
            try:
                line = int(attrs["line"])
            except ValueError, ex:
                raise CIXError("illegal <function> 'line' attribute: must "
                               "be integer: %r" % attrs["line"])
        else:
            line = None
        if attrs.has_key("lineend"):
            try:
                lineend = attrs["lineend"]
            except ValueError, ex:
                raise CIXError("illegal <function> 'lineend' attribute: must "
                               "be integer: %r" % attrs["lineend"])
        else:
            lineend = None
        if attrs.has_key("attributes"):
            attributes = attrs["attributes"]
        else:
            attributes = None
        if attrs.has_key("signature"):
            signature = attrs["signature"]
        else:
            signature = None
        type_ = ST_FUNCTION
        doc = None
        self.cu.execute("""INSERT INTO symbol
                            (file_id, scan_id, module_id, scope_id, scope,
                             name, signature, line, lineend, type, attributes,
                             doc)
                           VALUES (:file_id, :scan_id, :module_id,
                                   :scope_id, :scope, :name,
                                   :signature, :line, :lineend,
                                   :type_, :attributes, :doc)""",
                        locals())
        self.scope_id_stack.append(self.cx.last_insert_rowid())
        self.scope_stack.append(name)
    def end_function(self):
        self.scope_id_stack.pop()
        self.scope_stack.pop()

    def start_class(self, attrs):
        """<class .../>
        
        Attributes:
            "name" (required)
            "line" (optional) is a 1-based line number on which the
                definition begins. This is only really optional to support
                CIX data for binary files where a line number doesn't
                make sense. For scanned files a line should really be
                specified.
            "lineend" (optional, though strongly recommended) is a 1-based
                line number for the last line on which the definition occurs.
            "attributes" (optional, default null) is a space-separated list
                of attributes modifying this class.
            "signature" (optional, string)
        """
        file_id = self.file_id
        scan_id = self.scan_id
        module_id = self.module_id
        scope_id = self.scope_id_stack[-1] or 0
        scope = '.'.join(self.scope_stack)
        try:
            name = attrs["name"]
        except KeyError, ex:
            raise CIXError("<class> tag has no 'name' attribute")
        if attrs.has_key("line"):
            try:
                line = int(attrs["line"])
            except ValueError, ex:
                raise CIXError("illegal <class> 'line' attribute: must "
                               "be integer: %r" % attrs["line"])
        else:
            line = None
        if attrs.has_key("lineend"):
            try:
                lineend = attrs["lineend"]
            except ValueError, ex:
                raise CIXError("illegal <class> 'lineend' attribute: must "
                               "be integer: %r" % attrs["lineend"])
        else:
            lineend = None
        if attrs.has_key("attributes"):
            attributes = attrs["attributes"]
        else:
            attributes = None
        # The signature <class> attribute is currently deprecated.
        if attrs.has_key("signature"):
            signature = attrs["signature"]
        else:
            signature = None
        type_ = ST_CLASS
        doc = None
        self.cu.execute("""INSERT INTO symbol
                            (file_id, scan_id, module_id, scope_id, scope,
                             name, signature, line, lineend, type, attributes,
                             doc)
                           VALUES (:file_id, :scan_id, :module_id,
                                   :scope_id, :scope, :name,
                                   :signature, :line, :lineend,
                                   :type_, :attributes, :doc)""",
                        locals())
        self.scope_id_stack.append(self.cx.last_insert_rowid())
        self.scope_stack.append(name)
    def end_class(self):
        self.scope_id_stack.pop()
        self.scope_stack.pop()

    def start_interface(self, attrs):
        """<interface .../>
        
        Attributes:
            "name" (required)
            "line" (optional) is a 1-based line number on which the
                definition begins. This is only really optional to support
                CIX data for binary files where a line number doesn't
                make sense. For scanned files a line should really be
                specified.
            "lineend" (optional, though strongly recommended) is a 1-based
                line number for the last line on which the definition occurs.
            "attributes" (optional, default null) is a space-separated list
                of attributes modifying this class.
            "signature" (optional, string)
        """
        file_id = self.file_id
        scan_id = self.scan_id
        module_id = self.module_id
        scope_id = self.scope_id_stack[-1] or 0
        scope = '.'.join(self.scope_stack)
        try:
            name = attrs["name"]
        except KeyError, ex:
            raise CIXError("<interface> tag has no 'name' attribute")
        if attrs.has_key("line"):
            try:
                line = int(attrs["line"])
            except ValueError, ex:
                raise CIXError("illegal <interface> 'line' attribute: must "
                               "be integer: %r" % attrs["line"])
        else:
            line = None
        if attrs.has_key("lineend"):
            try:
                lineend = attrs["lineend"]
            except ValueError, ex:
                raise CIXError("illegal <interface> 'lineend' attribute: must "
                               "be integer: %r" % attrs["lineend"])
        else:
            lineend = None
        if attrs.has_key("attributes"):
            attributes = attrs["attributes"]
        else:
            attributes = None
        if attrs.has_key("signature"):
            signature = attrs["signature"]
        else:
            signature = None
        type_ = ST_INTERFACE
        doc = None
        self.cu.execute("""INSERT INTO symbol
                            (file_id, scan_id, module_id, scope_id, scope,
                             name, signature, line, lineend, type, attributes,
                             doc)
                           VALUES (:file_id, :scan_id, :module_id,
                                   :scope_id, :scope, :name,
                                   :signature, :line, :lineend,
                                   :type_, :attributes, :doc)""",
                        locals())
        self.scope_id_stack.append(self.cx.last_insert_rowid())
        self.scope_stack.append(name)
    def end_interface(self):
        self.scope_id_stack.pop()
        self.scope_stack.pop()

    def start_variable(self, attrs):
        """<variable .../>
        
        Attributes:
            "name" (required)
            "line" (optional) is a 1-based line number on which the
                definition begins. This is only really optional to support
                CIX data for binary files where a line number doesn't
                make sense. For scanned files a line should really be
                specified.
            "lineend" (optional, though strongly recommended) is a 1-based
                line number for the last line on which the definition occurs.
            "attributes" (optional, default null) is a space-separated list
                of attributes modifying this variable -- for example,
                "virtual public".
            "signature" (optional, string)
        """
        file_id = self.file_id
        scan_id = self.scan_id
        module_id = self.module_id
        scope_id = self.scope_id_stack[-1] or 0
        scope = '.'.join(self.scope_stack)
        try:
            name = attrs["name"]
        except KeyError, ex:
            raise CIXError("<variable> tag has no 'name' attribute")
        if attrs.has_key("line"):
            try:
                line = int(attrs["line"])
            except ValueError, ex:
                raise CIXError("illegal <variable> 'line' attribute: must "
                               "be integer: %r" % attrs["line"])
        else:
            line = None
        if attrs.has_key("attributes"):
            attributes = attrs["attributes"]
        else:
            attributes = None
        if attrs.has_key("signature"):
            signature = attrs["signature"]
        else:
            signature = None
        type_ = ST_VARIABLE
        doc = None
        self.cu.execute("""INSERT INTO symbol
                            (file_id, scan_id, module_id, scope_id, scope,
                             name, signature, line, type, attributes, doc)
                           VALUES (:file_id, :scan_id, :module_id,
                                   :scope_id, :scope, :name,
                                   :signature, :line, :type_,
                                   :attributes, :doc)""",
                        locals())
        last_symbol_id = self.cx.last_insert_rowid()
        self.scope_id_stack.append(last_symbol_id)
        self.scope_stack.append(name)
        self.last_typeable_row = ("symbol", last_symbol_id)
    def end_variable(self):
        self.scope_id_stack.pop()
        self.scope_stack.pop()
        self.last_typeable_row = (None, None)

    def start_argument(self, attrs):
        """<argument .../> - a function argument
        
        Attributes:
            "name" (required)
            "line" (optional) is a 1-based line number on which the argument
                definition begins
            "attributes" (optional, default null) is a space-separated list
                of attributes modifying this argument -- for example,
                "virtual public".
            "signature" (optional, string)
        """
        file_id = self.file_id
        scan_id = self.scan_id
        module_id = self.module_id
        scope_id = self.scope_id_stack[-1] or 0
        scope = '.'.join(self.scope_stack)
        try:
            name = attrs["name"]
        except KeyError, ex:
            raise CIXError("<argument> tag has no 'name' attribute")
        if attrs.has_key("line"):
            try:
                line = int(attrs["line"])
            except ValueError, ex:
                raise CIXError("illegal <argument> 'line' attribute: must "
                               "be integer: %r" % attrs["line"])
        else:
            line = None
        if attrs.has_key("attributes"):
            attributes = attrs["attributes"]
        else:
            attributes = None
        if attrs.has_key("signature"):
            signature = attrs["signature"]
        else:
            signature = None
        type_ = ST_ARGUMENT
        doc = None
        self.cu.execute("""INSERT INTO symbol
                            (file_id, scan_id, module_id, scope_id, scope,
                             name, signature, line, type, attributes, doc)
                           VALUES (:file_id, :scan_id, :module_id,
                                   :scope_id, :scope, :name,
                                   :signature, :line, :type_,
                                   :attributes, :doc)""",
                        locals())
        last_symbol_id = self.cx.last_insert_rowid()
        self.scope_id_stack.append(last_symbol_id)
        self.scope_stack.append(name)
        self.last_typeable_row = ("symbol", last_symbol_id)
    def end_argument(self):
        self.scope_id_stack.pop()
        self.scope_stack.pop()
        self.last_typeable_row = (None, None)

    def start_classref(self, attrs):
        """<classref .../>, i.e. a <class> base
        
        Attributes:
            "name" (required)
        """
        symbol_id = self.scope_id_stack[-1]
        scan_id = self.scan_id
        try:
            name = attrs["name"]
        except KeyError, ex:
            raise CIXError("<classref> tag has no 'name' attribute")
        type_ = BT_CLASSREF
        self.cu.execute("INSERT INTO base (symbol_id, scan_id, name, type) "
                        "VALUES (:symbol_id, :scan_id, :name, :type_)",
                        locals())
        self.last_typeable_row = ("base", self.cx.last_insert_rowid())
    def end_classref(self):
        self.last_typeable_row = (None, None)

    def start_interfaceref(self, attrs):
        """<interfaceref .../>, i.e. a <class> interface or <interface> base
        
        Attributes:
            "name" (required)
        """
        symbol_id = self.scope_id_stack[-1]
        try:
            name = attrs["name"]
        except KeyError, ex:
            raise CIXError("<interfaceref> tag has no 'name' attribute")
        type_ = BT_INTERFACEREF
        self.cu.execute("INSERT INTO base (symbol_id, name, type) "
                        "VALUES (:symbol_id, :name, :type_)",
                        locals())
        self.last_typeable_row = ("base", self.cx.last_insert_rowid())
    def end_interfaceref(self):
        self.last_typeable_row = (None, None)

    def start_type(self, attrs):
        """<type .../>, i.e. a possible type for a <variable/>, <argument/>
        <classref/> or <interfaceref/>.
        
        Attributes:
            "type" (required) is a CITDL expression
            "score" (optional) is a numeric value that _may_ be used to
                indicate higher probabilities of some type guesses over
                others.
        XXX Could add an optional line number that could be useful for
            showing how the type determination was made.
        """
        if self.lsf_language not in self.manager.coreLanguages:
            # OPTIMIZATION: Don't bother collecting type information for
            # languages that don't use CI-based completion (the only use
            # for the type info).
            return
        
        scan_id = self.scan_id
        table, target_id = self.last_typeable_row
        if table is None or target_id is None:
            raise CIXError("<type> tag has no appropriate target symbol or base")
        if table == "symbol":
            symbol_id = target_id
            base_id = None
        elif table == "base":
            symbol_id = None
            base_id = target_id
        else:
            raise CIXError("unexpected target table for <type> tag: '%s'" % table)
        try:
            citdl = attrs["type"]
        except KeyError, ex:
            raise CIXError("<type> tag has no 'type' attribute")
        if attrs.has_key("score"):
            try:
                score = int(attrs["score"])
            except ValueError, ex:
                raise CIXError("optional 'score' attribute of <type> tag "
                               "must be an integer: %r" % attrs["score"])
        else:
            score = None
        self.cu.execute("""INSERT INTO type
                            (scan_id, symbol_id, base_id, citdl, score)
                           VALUES (:scan_id, :symbol_id, :base_id,
                                   :citdl, :score)""",
                        locals())

    def start_import(self, attrs):
        """<import .../>
        
        Attributes:
            "module" (required) is the name of the module (or file) being
                imported.
            "symbol" (optional) is the name of a symbol being imported from
                'module', if one is specified. E.g.:
                    Python: from time import asctime
                    Perl:   use File::Path qw(mkpath);
            "alias" (optional) is an alias for 'name' if any, or for 'module',
                if not. E.g.
                    Python:     import xml as XML
                    Python:     from time import time as clock
            "line" (optional) is a 1-based line number on which the import
                definition begins
        """
        file_id = self.file_id
        scan_id = self.scan_id
        module_id = self.module_id
        # We can't use None/NULL as the scope_id to indicate "module-level
        # scope" because you can't SELECT on that, at least not with SQLite.
        scope_id = self.scope_id_stack[-1] or 0
        try:
            module = attrs["module"]
        except KeyError, ex:
            raise CIXError("<import> tag has no 'module' attribute")
        try:
            line = int(attrs["line"])
        except KeyError, ex:
            line = None
        except ValueError, ex:
            raise CIXError("illegal <import> 'line' attribute: must "
                           "be an integer: %r" % attrs["line"])
        if attrs.has_key("symbol"):
            symbol = attrs["symbol"]
        else:
            symbol = None
        if attrs.has_key("alias"):
            alias = attrs["alias"]
        else:
            alias = None
        self.cu.execute("""INSERT INTO import
                            (file_id, scan_id, module_id, scope_id, line,
                             module, symbol, alias)
                           VALUES (:file_id, :scan_id, :module_id,
                                   :scope_id, :line, :module,
                                   :symbol, :alias)""",
                        locals())



#---- public classes

class Database:
    """A friendlier interface to a Code Intelligence database (CIDB).
    
    Usage:
        db = Database(manager)
        print db.getVersion()
    
    Note that you may only use a Database from one thread because a CI
    database connection cannot be shared by multiple threads.
    """
    CIDB_VERSION = (1, 0)

    def __init__(self, manager):
        """Create a Database instance.
        
            "manager" is the controlling codeintel.Manager instance.
        """
        self.manager = manager

    def create(self, dbPath=None):
        """Setup the database for first use."""
        # Create the tables.
        fname = os.path.join(os.path.dirname(__file__), "cidb_schema.sql")
        schema = open(fname, "r").read()
        cx = self.manager.getDatabaseConnection(dbPath)
        cu = cx.cursor()
        self.manager.acquireCIDBWriteLock()
        try:
            cu.execute("PRAGMA synchronous=OFF;")
            cu.execute(schema)
            
            # Setup some "meta" table information.
            ver = '.'.join(map(str, self.CIDB_VERSION))
            cu.execute("BEGIN")
            cu.execute("INSERT INTO meta(name, value) "
                       "VALUES ('version', ?)", (ver,))
            cu.execute("INSERT INTO meta(name, value) "
                       "VALUES ('hostname', ?)", (socket.gethostname(),))
            cu.execute("INSERT INTO meta(name, value) "
                       "VALUES ('platform', ?)", (getPlatform(),))
            cu.execute("COMMIT")
    
            # Load stdcix files for core langages.
            stdcixDir = os.path.dirname(__file__)
            for lang in self.manager.coreLanguages:
                src = os.path.join(stdcixDir, lang.lower()+".cix")
                if not os.path.exists(src):
                    continue
                self._upgrade_stdcix_file(cx, cu, None, src, None)
        finally:
            self.manager.releaseCIDBWriteLock()
            cu.finalize()
            cx.close()

    def _isSQLite2(self, cu, dbPath):
        """Return true if the given CIDB path is of sqlite2 binary format.
        
            "cu" is an open APSW/sqlite3 cursor.
            "dbPath" is the path to the database being checked.
        """
        try:
            cu.execute("SELECT * FROM meta LIMIT 1")
        except apsw.NotADBError:
            # apsw.NotADBError: NotADBError: file is encrypted or is not a database
            import sqlite
            try:
                cx2 = sqlite.Connection(dbPath, autocommit=1)
            except sqlite.Error:
                pass # that CIDB is just corrupt
            else:
                # The CIDB is of the sqlite2 binary format. This must be
                # upgraded.
                cx2.close()
                return True
        return False

    def getUpgradeInfo(self, dbPath, minVersion):
        """Return a 2-tuple indicating if this database requires an upgrade.
        
            "dbPath" is the full path the the CIDB.
            "minVersion" is the minimum CIDB version required.
        
        Returns one of the following:
            (CIDB_UPGRADE_NOT_NECESSARY, None)
            (CIDB_UPGRADE_NOT_POSSIBLE, "<reason>")
            (CIDB_UPGRADE_NECESSARY, None)
            
        Users should generally just call Manager.getCIDBUpgradeInfo().
        """
        # If the database is still the old sqlite2 binary format, we need
        # to upgrade it.
        cx = self.manager.getDatabaseConnection(dbPath)
        cu = cx.cursor()
        self.manager.acquireCIDBReadLock()
        try:
            if self._isSQLite2(cu, dbPath):
                return (CIDB_UPGRADE_NECESSARY, None)
            
            if self._getHostname(cu) != socket.gethostname():
                return (CIDB_UPGRADE_NOT_POSSIBLE,
                        "database '%s' hostname, '%s', does not match the "
                        "current hostname: '%s'"
                        % (dbPath, self.getHostname(), socket.gethostname()))
            if self._getPlatform(cu) != getPlatform():
                return (CIDB_UPGRADE_NOT_POSSIBLE,
                        "database '%s' platform, '%s', does not match "
                        "the current platform: '%s'"
                        % (dbPath, self.getPlatform(), getPlatform()))
            verInfo = self._getVersionInfo(cu)
            if verInfo < (0, 14):
                return (CIDB_UPGRADE_NOT_POSSIBLE,
                        "database '%s' is of version %s.%s: do not know how to "
                        "upgrade pre-0.14 databases"
                        % (dbPath, verInfo[0], verInfo[1]))
            if verInfo < minVersion:
                return (CIDB_UPGRADE_NECESSARY, None)
    
            # Upgrade if there are newer standard language CIX files than
            # those already loaded into the current CIDB or if the standard
            # CIX file for any of the "core" languages is not loaded.
            stdcix2mtime = {}
            for lang in self.manager.coreLanguages:
                stdcix2mtime[lang.lower()+".cix"] = (None, None)
            cu.execute("SELECT file.compare_path,scan.mtime FROM scan,file "
                       "WHERE scan.generator='StdCIX' "
                       "AND file.id=scan.file_id")
            for file_cpath, scan_mtime in cu:
                scan_mtime = int(scan_mtime) # PySQLite doesn't always convert
                stdcix2mtime[os.path.basename(file_cpath)] = (file_cpath, scan_mtime)
    
            stdcixDir = os.path.dirname(__file__)
            mtimePat = re.compile(r'<file .*?mtime="(.*?)"', re.S)
            for stdcixBase, (scan_path, scan_mtime) in stdcix2mtime.items():
                src = os.path.join(stdcixDir, stdcixBase)
                if not os.path.exists(src):
                    continue
    
                if scan_mtime is None: # core stdcix file that is not loaded
                    return (CIDB_UPGRADE_NECESSARY, None)
                else:
                    # Extract 'mtime' attr from <file> tag.
                    try:
                        fin = open(src, 'r')
                        try:
                            head = fin.read(1024)
                        finally:
                            fin.close()
                        src_mtime = int(mtimePat.search(head).group(1))
                    except (EnvironmentError, ValueError, KeyError), ex:
                        log.warn("could not determine if '%s' needs "
                                 "updating: %s", stdcixbase, ex)
                        continue
                    if src_mtime > scan_mtime:
                        log.info("upgrade necessary: mtime for '%s' in "
                                 "database (%s) is less than on disk (%s)",
                                 stdcixBase, scan_mtime, src_mtime)
                        return (CIDB_UPGRADE_NECESSARY, None)
    
            return (CIDB_UPGRADE_NOT_NECESSARY, None)
        finally:
            self.manager.releaseCIDBReadLock()
            cu.finalize()
            cx.close()

    def _upgrade_to_sqlite3(self, dbPath, progressCB=None):
        """Upgrade the CIDB from sqlite2 binary format for sqlite3.
        
        Raises CIDBError if could not upgrade.
        """
        if progressCB:
            stage = "upgrading to binary format 3: preparing..."
            progressCB(stage, 1)

        try:
            import sqlite
            assert sqlite.apilevel.startswith("2.")
        except (ImportError, AssertionError), ex:
            raise CodeIntelError("cannot upgrade '%s': don't have PySQLite for sqlite2" % dbpath)
        try:
            import apsw
            assert apsw.sqlitelibversion().startswith("3.")
        except (ImportError, AssertionError), ex:
            raise CodeIntelError("cannot upgrade '%s': don't have apsw for sqlite3" % dbpath)
        tmpPath = dbPath+".tmp"
        if os.path.exists(tmpPath):
            os.remove(tmpPath)
        journalPath = tmpPath+"-journal"
        if os.path.exists(journalPath):
            # A journal from a previous attempt (which presumably
            # crashed or was killed) can cause problems.
            os.remove(journalPath)
        bakPath = dbPath+".bak"
    
        try:
            log.debug("upgrading '%s' from sqlite2 to sqlite3..." % dbPath)
            cx2 = sqlite.Connection(dbPath, autocommit=1)
            cu2 = cx2.cursor()
            cx3 = apsw.Connection(tmpPath)
            cu3 = cx3.cursor()
            
            try:
                log.info("create CIDB schema in '%s'", tmpPath)
                if progressCB:
                    stage = "upgrading to binary format 3: create schema"
                    progressCB(stage, 2)
                cu3.execute("BEGIN")
                cu2.execute("SELECT type, name, sql FROM sqlite_master")
                tables = []
                for (type, name, sql) in cu2.fetchall():
                    # With APSW's more strict handling of sqlite type
                    # conversions the exact text (case and name) of declared
                    # column types matter more. In particular: lowercase and
                    # "unicode" are useless. That was just a PySQLite-thang.
                    if type == "table":
                        sql = sql.replace("text", "TEXT")
                        sql = sql.replace("unicode", "TEXT")
                        sql = sql.replace("integer", "INTEGER")
                    cu3.execute(sql)
                    if type == "table":
                        tables.append(name)
                cu3.execute("COMMIT")
                
                # Unicode woes:
                # I think either PySQLite or sqlite2 was screwing up here
                # because I get the following in the next block with Shane's
                # "0.14/shanec/shanec2-codeintel.db":
                #   UnicodeDecodeError: 'ascii' codec can't decode byte 0xe7 in position 69: ordinal not in range(128)
                # For this row of data:
                #   (476, '/home/shanec/src/egroupware/addressbook/inc/export/Outlook_CSV_-_Fran\xe7ais', '/home/shanec/src/egroupware/addressbook/inc/export/Outlook_CSV_-_Fran\xe7ais', 5)
                # Other rows have unicode strings. Perhaps this is just an
                # acceptible SNAFU with changes in default file system
                # encoding? Encodings with which to attempt recovery:
                encodings = [sys.getfilesystemencoding(), "iso8859-15"]
                if sys.platform == "win32": encodings.append("mbcs")
                
                if progressCB:
                    BASE = 3
                    RANGES = {"meta": 1, "language": 1, "file": 1, "scan": 1,
                              "module": 2, "symbol": 70, "base": 3, "type": 14,
                              "import": 3}
                    progressCB(stage, BASE)
                cu3.execute("BEGIN")
                for i, table in enumerate(tables):
                    log.info("copy over '%s' table", table)
                    if progressCB:
                        stage = "upgrading to binary format 3: %s table" % table
                        RANGE = RANGES.get(table, 0)
                        SUBBASE = BASE
                        progressCB(stage, SUBBASE)
                        BASE += RANGE
                    cu2.execute("SELECT * FROM "+table)
                    rows = [tuple(row) for row in cu2.fetchall()]
                    if rows:
                        sql = "INSERT INTO %s VALUES(%s)"\
                              % (table, ','.join(['?']*len(rows[0])))
                        for j, row in enumerate(rows):
                            if progressCB and j%3000==0:
                                progressCB(stage, SUBBASE + j*RANGE/len(rows))
                            try:
                                cu3.execute(sql, row)
                            except UnicodeDecodeError:
                                for encoding in encodings:
                                    lrow = list(row)
                                    for i in range(len(lrow)):
                                        if isinstance(lrow[i], str):
                                            try:
                                                lrow[i] = lrow[i].decode(encoding)
                                            except UnicodeDecodeError:
                                                break
                                    cu3.execute(sql, lrow)
                                    break
                                else:
                                    raise
                            except apsw.ConstraintError, ex:
                                # Work around this unexpected problem with old CIDBs:
                                #   apsw.ConstraintError: ConstraintError: PRIMARY KEY must be unique
                                # One from JeffG (cidb_corpus/0.16/jeffg/codeintel.db)
                                # had bogus rows like this:
                                #   -500152526,None,None,None,None,None,None,None,None,None,None,None,None
                                # We just drop them.
                                pass
                cu3.execute("COMMIT")
            finally:
                cu2.close()
                cx2.close()
                del cu3
                del cx3
        except (sqlite.Error, apsw.Error), ex:
            raise CIDBError("could not upgrade database: %s" % ex)
    
        log.info("create backup of original at '%s'", bakPath)
        if progressCB:
            stage = "upgrading to binary format 3: backup original"
            progressCB(stage, 99)
        if os.path.exists(bakPath):
            os.remove(bakPath)
        shutil.copy(dbPath, bakPath)
        shutil.copy(tmpPath, dbPath)
        os.remove(tmpPath)

    def _upgrade_stdcix_file(self, cx, cu, oldPath, newPath, progressCB=None):
        """Upgrade the given stdcix file.
        
            "oldPath" is the old stdcix file in the database or None, if
                one isn't loaded.
            "newPath" is the new stdcix file to load.
            "progressCB" (optional) is a callback as per upgrade()'s doc.
        """
        log.debug("_upgrade_stdcix_file(cx, cu, oldPath=%r, newPath=%r)",
                  oldPath, newPath)
        if oldPath:
            stage = "upgrade standard '%s'" % os.path.basename(newPath)
        else:
            stage = "loading standard '%s'" % os.path.basename(newPath)
        if progressCB: progressCB(stage, 0)

        if oldPath:
            if 1:
                self._removeFile(cu, oldPath, commit=False)
            else:
                self.removeFile(oldPath, commit=False)
        if progressCB: progressCB(stage, 30)

        f = open(newPath, 'r')
        try:
            content =  f.read()
        finally:
            f.close()
        self.update(content, recover=False, scan_imports=False, cx=cx, cu=cu)
        if progressCB: progressCB(stage, 100)

    def _upgrade_core_stdcix(self, cx, cu, dbPath, progressCB=None):
        # Upgrade if there are newer standard language CIX files than those
        # already loaded into the current CIDB or if the standard CIX file
        # for any of the "core" languages is not loaded.
        cu.execute("PRAGMA synchronous=OFF")
        stdcix2mtime = {}
        for lang in self.manager.coreLanguages:
            stdcix2mtime[lang.lower()+".cix"] = (None, None)
        cu.execute("SELECT file.id,file.compare_path,scan.mtime "
                   "FROM scan,file WHERE scan.generator='StdCIX' "
                   "AND file.id=scan.file_id")
        for file_id, file_cpath, scan_mtime in tuple(cu):
            if sys.platform != "win32"\
               and os.path.basename(file_cpath).startswith("codeintel\\"):
                # Fix a SNAFU where python.cix and perl.cix included
                # a Windows-specific path separator that got loaded
                # into the CIDB on Unix, there by screwing up this
                # new stdcix upgrading.
                log.debug("remove stdcix SNAFU file: %r" % file_cpath)
                cu.execute("BEGIN")
                try:
                    self._removeFileId(cu, file_id)
                except:
                    cu.execute("ROLLBACK")
                    raise
                else:
                    cu.execute("COMMIT")
                continue
            scan_mtime = int(scan_mtime) # PySQLite doesn't always convert
            stdcix2mtime[os.path.basename(file_cpath)] = (file_cpath, scan_mtime)

        stdcixDir = os.path.dirname(__file__)
        mtimePat = re.compile(r'<file .*?mtime="(.*?)"', re.S)
        for stdcixBase, (scan_path, scan_mtime) in stdcix2mtime.items():
            src = os.path.join(stdcixDir, stdcixBase)
            if not os.path.exists(src):
                continue
            
            if scan_path is None: # core stdcix file that is not loaded
                self._upgrade_stdcix_file(cx, cu, None, src, progressCB)
            else:
                # Extract 'mtime' attr from <file> tag.
                try:
                    fin = open(src, 'r')
                    try:
                        head = fin.read(1024)
                    finally:
                        fin.close()
                    src_mtime = int(mtimePat.search(head).group(1))
                except (EnvironmentError, ValueError, KeyError), ex:
                    log.warn("could not determine if '%s' needs "
                             "updating: %s", stdcixbase, ex)
                    continue
                if src_mtime > scan_mtime:
                    self._upgrade_stdcix_file(cx, cu, scan_path, src, progressCB)

    def _upgrade_014_015(self, cu, dbPath, progressCB=None):
        """Upgrade from CIDB v0.14 to CIDB v0.15"""
        if progressCB:
            stage = "upgrading from CIDB v0.14 to v0.15"
            progressCB(stage, 0)
        cu.execute("PRAGMA synchronous=OFF; BEGIN")

        try:
            # Gather data for upgrading
            lang2lang_id = {}
            lang_id2lang = {}
            for lang_id, lang in cu.execute("SELECT id,name FROM language"):
                lang_id2lang[lang_id] = lang
            lang2file_ids = {}
            if progressCB: progressCB(stage, 5)
            for file_id, lang_id in cu.execute("SELECT id,language_id FROM file"):
                lang = lang_id2lang[lang_id]
                if lang not in lang2file_ids:
                    lang2file_ids[lang] = []
                lang2file_ids[lang].append(file_id)
            if progressCB: progressCB(stage, 10)
    
            pyfile_ids = lang2file_ids.get("Python", [])
            # - Python: drop argument signatures
            # These were dropped because they are no longer used for calltip
            # construction.
            BASE, RANGE = 10, 20
            for i, file_id in enumerate(pyfile_ids):
                if progressCB and i % 20 == 0:
                    progressCB(stage, BASE + i*RANGE/len(pyfile_ids))
                cu.execute("UPDATE symbol SET signature=? WHERE type=? AND file_id=?",
                           (None, ST_ARGUMENT, file_id))
            if progressCB: progressCB(stage, 30)
            # - Python: regen function & class sig and doc from existing doc
            # - Python: drop leading "def" on function signatures (if there is one)
            # - Python: trim doc strings for all other symbol types
            BASE, RANGE = 30, 69
            for i, file_id in enumerate(pyfile_ids):
                cu.execute("SELECT id, type, signature, doc FROM symbol WHERE file_id=%d" % file_id)
                for symbol_id, type_, signature, doc in tuple(cu):
                    if type_ == ST_FUNCTION:
                        changed = False
                        siglines, doclines = parsePyFuncDoc(doc)
                        if not siglines:
                            if signature and signature.startswith("def "):
                                signature = signature[4:]
                                changed = True
                        else:
                            signature = '\n'.join(siglines)
                            changed = True
                        if doclines:
                            doc = '\n'.join(doclines)
                            changed = True
                        if changed:
                            cu.execute("UPDATE symbol SET signature=?, doc=? WHERE id=?",
                                       (signature, doc, symbol_id))
                    elif type_ == ST_CLASS:
                        siglines, doclines = parsePyFuncDoc(doc)
                        signature = '\n'.join(siglines)
                        if doclines:
                            doc = '\n'.join(doclines)
                        cu.execute("UPDATE symbol SET signature=?, doc=? WHERE id=?",
                                   (signature, doc, symbol_id))
                    elif doc:
                        doclines = parseDocSummary(doc.splitlines(0))
                        cu.execute("UPDATE symbol SET doc=? WHERE id=?",
                                   ('\n'.join(doclines), symbol_id))
                cu.execute("SELECT id, doc FROM module WHERE file_id=%d" % file_id)
                for module_id, doc in cu:
                    if doc:
                        doclines = parseDocSummary(doc.splitlines(0))
                        cu.execute("UPDATE module SET doc=? WHERE id=?",
                                   ('\n'.join(doclines), module_id))
                if progressCB and i % 20 == 0:
                    progressCB(stage, BASE + i*RANGE/len(pyfile_ids))
            if progressCB: progressCB(stage, 99)
    
            cu.execute("UPDATE meta SET value='0.15' WHERE name='version'")
        except:
            cu.execute("ROLLBACK")
            raise
        else:
            cu.execute("COMMIT")

    def _upgrade_015_016(self, cu, dbPath, progressCB=None):
        """Upgrade to CIDB v0.16 from v0.15
        
        Unfortunately the Perl scanning has changed significantly enought
        that all those files just need to be re-scanned. The "upgrade" here
        is to just ditch all the Perl scans.
        """
        if progressCB:
            stage = "upgrading from CIDB v0.15 to v0.16"
            progressCB(stage, 0)
        cu.execute("PRAGMA synchronous=OFF; BEGIN")
        try:
            cu.execute("SELECT id FROM language WHERE name='Perl' LIMIT 1")
            try:
                perllang_id = cu.next()[0]
                if progressCB: progressCB(stage, 5)
    
                rows = tuple(cu.execute(
                    "SELECT id FROM file WHERE language_id=%d" % perllang_id))
                if progressCB: progressCB(stage, 10)
                BASE, RANGE = 10, 89
                nfiles = len(rows)
                for i, row in enumerate(rows):
                    if progressCB and i % 20 == 0:
                        progressCB(stage, BASE + i*RANGE/nfiles)
                    self._removeFileId(cu, row[0])
            except StopIteration:
                pass
    
            if progressCB: progressCB(stage, 99)
            cu.execute("UPDATE meta SET value='0.16' WHERE name='version'")
        except:
            cu.execute("ROLLBACK")
            raise
        else:
            cu.execute("COMMIT")

    def _upgrade_016_100(self, cu, dbPath, progressCB=None):
        """Upgrade to CIDB v1.0 from v0.16
        
        This is a _little_ bit of a hack because the 0.16->1.0 step is
        actually a change of the CIDB schema column _type_ names to the
        strict set that sqlite (and APSW) recognize -- rather than the
        wider set that PySQLite works with. The actual schema convertion
        is handled by _upgrade_to_sqlite3(), which is run a little earlier.
        """
        if progressCB:
            stage = "upgrading from CIDB v0.16 to v1.0"
            progressCB(stage, 20)
        if progressCB: progressCB(stage, 90)
        cu.execute("UPDATE meta SET value='1.0' WHERE name='version'")

    _upgraderMap = {
        (0,14): {(0,15): _upgrade_014_015},
        (0,15): {(0,16): _upgrade_015_016},
        (0,16): {(1,0): _upgrade_016_100},
    }

    
    def upgrade(self, dbPath, progressCB=None, mode=None):
        """Upgrade the database to the current CIDB version.
        
            "progressCB" (optional) is a callback with which the upgrade
                can report progress. It should be called with two arguments:
                "stage", a string describing the current upgrade stage; and
                "percent", an integer value between 0 and 100 indicating
                estimated percent completion.
            "mode" (optional, default "restore") specifies the failure mode.
                It must be either "restore" to restore the old DB on failure
                or "replace" to replace the old DB with a new one on failure.
        
        Raises CIDBError if the upgrade fails.
        """
        try:
            cx = self.manager.getDatabaseConnection(dbPath)
            cu = cx.cursor()
            self.manager.acquireCIDBWriteLock()
            try:
                if self._isSQLite2(cu, dbPath):
                    self._upgrade_to_sqlite3(dbPath, progressCB)
            finally:
                self.manager.releaseCIDBWriteLock()
                cu.finalize()
                cx.close()
    
            cx = self.manager.getDatabaseConnection(dbPath)
            cu = cx.cursor()
            self.manager.acquireCIDBWriteLock()
            try:
                def joinver(ver):
                    return '.'.join(map(str, ver))
                currVer = startVer = self._getVersionInfo(cu)
                targetVer = self.CIDB_VERSION
                while currVer != targetVer:
                    try:
                        availUpgraderMap = self._upgraderMap[currVer]
                        availUpgradeTos = availUpgraderMap.keys()
                        availUpgradeTos.sort()
                        upgradeTo = availUpgradeTos[-1]
                        upgrader = availUpgraderMap[availUpgradeTos[-1]]
                    except KeyError:
                        raise CIDBError("cannot upgrade from CIDB v%s: no upgrader "
                                        "for this version" % joinver(currVer))
                    log.info("upgrading from CIDB v%s to CIDBv%s ...",
                             joinver(currVer), joinver(upgradeTo))
                    upgrader(self, cu, dbPath, progressCB)
                    currVer = upgradeTo

                self._upgrade_core_stdcix(cx, cu, dbPath, progressCB)
            finally:
                self.manager.releaseCIDBWriteLock()
                cu.finalize()
                cx.close()
        except CodeIntelError, ex:
            if mode == "replace":
                self.manager.resetCIDB(dbPath)
                self.create(dbPath)
            raise

    def update(self, cix, request=None, driver=None, recover=True,
               scan_imports=True, cx=None, cu=None):
        """Update the database with the given CIX-format results.
        
            "cix" is the XML content of the scan results for the given
                request. The XML must be in the std Code Intelligence XML
                format.
            "request" (optional) is the ScanRequest instance that resulted in
                the given CIX.
            "driver" (optional) is the LanguageEngineDriver instance that
                was used to generate the given CIX.
            "recover" is an internal boolean for recursive calls indicating
                if error recovery should be attempted. It is only attempted
                on the first time through.
            "scan_imports" (optional, default true) is a boolean indicating
                if the processing of file imports should be scheduled.
            "cx", "cu" (optional) are a borrowed CIDB connection and cursor
                to use. If not specified, a database connection and cursor
                are opened.

        Can raise a CodeIntelError if there is a problem.
        """
        # xml.sax expects an encoded string. Ensure that that's what
        # 'cix' is.
        if isinstance(cix, unicode):
            encoding = _detectXMLEncoding(cix)
            if not encoding:
                encoding = "utf-8" # fallback to trying UTF-8
            try:
                cix = cix.encode(encoding)
            except UnicodeError, ex:
                raise CIXError("error updating database with CIX: could not "
                               "encoding the unicode CIX content: "
                               "encoding='%s'" % encoding)

        if cx is not None and cu is not None:
            borrowedConnection = True
        else:
            borrowedConnection = False
            cx = self.manager.getDatabaseConnection()
            cu = cx.cursor()
            self.manager.acquireCIDBWriteLock()
        try:
            cu.execute("PRAGMA synchronous=OFF; BEGIN")
            ch = _Cix2CidbHandler(cx, cu, self.manager,
                                  scan_imports=scan_imports)
            try:
                xml.sax.parseString(cix, ch)
            except (CIXError, xml.sax.SAXException, apsw.Error,
                    UnicodeError), ex:
                cu.execute("ROLLBACK") # Drop any changes that *were* made.
                if not recover:
                    raise
                # Try to recover.
                if ch.lsf_path: # means parsing got as far as a <file/>
                    fattrs = {"path": ch.lsf_path,
                              "generator": ch.lsf_generator,
                              "language": ch.lsf_language,
                              "error": str(ex)}
                    cix2 = '<?xml version="1.0" encoding="UTF-8"?>'\
                           '<codeintel version="0.1"> <file%s/> </codeintel>'\
                           % xmlattrstr(fattrs)
                    return self.update(cix2, request, driver, recover=0,
                                       scan_imports=scan_imports,
                                       cx=cx, cu=cu)
                elif request and driver:
                    fattrs = {"path": request.path,
                              "generator": driver.name,
                              "language": request.language,
                              "error": str(ex)}
                    cix2 = '<?xml version="1.0" encoding="UTF-8"?>'\
                           '<codeintel version="0.1"> <file%s/> </codeintel>'\
                           % xmlattrstr(fattrs)
                    return self.update(cix2, request, driver, recover=0,
                                       scan_imports=scan_imports,
                                       cx=cx, cu=cu)
                else:
                    raise
            cu.execute("COMMIT")
        finally:
            if not borrowedConnection:
                self.manager.releaseCIDBWriteLock()
                cu.finalize()
                cx.close()

        # Update the relevant cache bits for this update.
        factory = self.manager.factory
        for file_id, file_path in ch.updated_files:
            factory.invalidateCache(file_id, file_path)
        # - This is a little bit of a HACK, in that this is hardcoding
        #   deep knowledge about the caching data structures.
        perlPkgInfoCache = self.manager.cache.get("perlPkgInfoCache", None)
        if perlPkgInfoCache:
            for pm in ch.updated_perl_modules:
                if pm in perlPkgInfoCache:
                    del perlPkgInfoCache[pm]

    def _removeFileId(self, cu, file_id):
        cu.execute("SELECT id FROM scan WHERE file_id=%d" % file_id)
        for (scan_id,) in tuple(cu):
            cu.execute("DELETE FROM module WHERE scan_id=%d" % scan_id)
            cu.execute("DELETE FROM symbol WHERE scan_id=%d" % scan_id)
            cu.execute("DELETE FROM base WHERE scan_id=%d" % scan_id)
            cu.execute("DELETE FROM import WHERE scan_id=%d" % scan_id)
            cu.execute("DELETE FROM type WHERE scan_id=%d" % scan_id)
            cu.execute("DELETE FROM scan WHERE id=%d" % scan_id)
        cu.execute("DELETE FROM file WHERE id=%d" % file_id)

    def _removeFile(self, cu, path, commit=True):
        cu.execute("PRAGMA synchronous=OFF")
        if commit: cu.execute("BEGIN")
        cpath = canonicalizePath(path)
        cu.execute("SELECT id FROM file WHERE compare_path=? LIMIT 1", (cpath,))
        try:
            file_id = int(cu.next()[0])
        except StopIteration:
            if commit: cu.execute("ROLLBACK")
            log.warn("could not remove file '%s': not in CIDB (cpath=%r)",
                     path, cpath)
        else:
            self._removeFileId(cu, file_id)
            if commit: cu.execute("COMMIT")

    def removeFile(self, path, commit=True):
        cx = self.manager.getDatabaseConnection()
        cu = cx.cursor()
        self.manager.acquireCIDBWriteLock()
        try:
            self._removeFile(cu, path, commit=commit)
        finally:
            self.manager.releaseCIDBWriteLock()
            cu.finalize()
            cx.close()

    def _getVersion(self, cu):
        cu.execute("SELECT name, value FROM meta WHERE name='version' LIMIT 1")
        return cu.next()[1]
    def getVersion(self):
        cx = self.manager.getDatabaseConnection()
        cu = cx.cursor()
        self.manager.acquireCIDBReadLock()
        try:
            return self._getVersion(cu)
        finally:
            self.manager.releaseCIDBReadLock()
            cu.finalize()
            cx.close()

    def _getVersionInfo(self, cu):
        return _version2versionInfo(self._getVersion(cu))
    def getVersionInfo(self):
        return _version2versionInfo(self.getVersion())

    def _getHostname(self, cu):
        cu.execute("SELECT name, value FROM meta WHERE name='hostname' LIMIT 1")
        return cu.next()[1]
    def getHostname(self):
        cx = self.manager.getDatabaseConnection()
        cu = cx.cursor()
        self.manager.acquireCIDBReadLock()
        try:
            return self._getHostname(cu)
        finally:
            self.manager.releaseCIDBReadLock()
            cu.finalize()
            cx.close()

    def _getPlatform(self, cu):
        cu.execute("SELECT name, value FROM meta WHERE name='platform'")
        return cu.next()[1]
    def getPlatform(self):
        cx = self.manager.getDatabaseConnection()
        cu = cx.cursor()
        self.manager.acquireCIDBReadLock()
        try:
            return self._getPlatform(cu)
        finally:
            self.manager.releaseCIDBReadLock()
            cu.finalize()
            cx.close()


